# pylint: disable=too-many-lines
from typing import (
    Optional,
    Sequence,
    Union,
)
from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.common.const import (
    PCMK_ROLE_STARTED,
    PCMK_ROLES,
    PCMK_STATUS_ROLE_PROMOTED,
    PCMK_STATUS_ROLE_STARTED,
    PCMK_STATUS_ROLE_STOPPED,
    PCMK_STATUS_ROLE_UNPROMOTED,
    PCMK_STATUS_ROLES,
    PCMK_STATUS_ROLES_PENDING,
    PCMK_STATUS_ROLES_RUNNING,
    PcmkStatusRoleType,
)
from pcs.common.status_dto import (
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)
from pcs.lib.pacemaker import status

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_equal,
    assert_report_item_list_equal,
)


def fixture_primitive_xml(
    resource_id: str = "resource",
    resource_agent: str = "ocf:heartbeat:Dummy",
    role: PcmkStatusRoleType = PCMK_STATUS_ROLE_STARTED,
    target_role: Optional[str] = None,
    managed: bool = True,
    node_names: Sequence[str] = ("node1",),
    add_optional_args: bool = False,
) -> str:
    target_role = (
        f'target_role="{target_role}"' if target_role is not None else ""
    )
    active = role in PCMK_STATUS_ROLES_RUNNING
    description = 'description="Test description"' if add_optional_args else ""
    pending = 'pending="test"' if add_optional_args else ""
    locked_to = 'locked_to="test"' if add_optional_args else ""

    nodes = "\n".join(
        f'<node name="{node}" id="{i}" cached="true"/>'
        for (i, node) in enumerate(node_names)
    )

    return f"""
        <resource
            id="{resource_id}"
            resource_agent="{resource_agent}"
            role="{role}"
            {target_role}
            active="{active}"
            orphaned="false"
            blocked="false"
            maintenance="false"
            {description}
            managed="{managed}"
            failed="false"
            failure_ignored="false"
            {pending}
            {locked_to}
            nodes_running_on="{len(node_names)}"
        >
            {nodes}
        </resource>
    """


def fixture_primitive_dto(
    resource_id: str = "resource",
    instance_id: Optional[str] = None,
    resource_agent: str = "ocf:heartbeat:Dummy",
    role: PcmkStatusRoleType = PCMK_STATUS_ROLE_STARTED,
    target_role: Optional[str] = None,
    managed: bool = True,
    node_names: Sequence[str] = ("node1",),
    add_optional_args: bool = False,
) -> PrimitiveStatusDto:
    return PrimitiveStatusDto(
        resource_id,
        instance_id,
        resource_agent,
        role,
        target_role,
        active=role in PCMK_STATUS_ROLES_RUNNING,
        orphaned=False,
        blocked=False,
        maintenance=False,
        description="Test description" if add_optional_args else None,
        managed=managed,
        failed=False,
        failure_ignored=False,
        node_names=list(node_names),
        pending="test" if add_optional_args else None,
        locked_to="test" if add_optional_args else None,
    )


def fixture_group_xml(
    resource_id: str = "resource-group",
    description: Optional[str] = None,
    members: Sequence[str] = (),
) -> str:
    description = (
        f'description="{description}"' if description is not None else ""
    )
    members = "\n".join(members)
    return f"""
        <group
            id="{resource_id}"
            number_resources="{len(members)}"
            maintenance="false"
            {description}
            managed="true"
            disabled="false"
        >
            {members}
        </group>
    """


def fixture_group_dto(
    resource_id: str = "resource-group",
    instance_id: Optional[str] = None,
    description: Optional[str] = None,
    members: Sequence[PrimitiveStatusDto] = (),
) -> GroupStatusDto:
    return GroupStatusDto(
        resource_id,
        instance_id,
        maintenance=False,
        description=description,
        managed=True,
        disabled=False,
        members=list(members),
    )


def fixture_clone_xml(
    resource_id: str = "resource-clone",
    multi_state: bool = False,
    unique: bool = False,
    description: Optional[str] = None,
    target_role: Optional[str] = None,
    instances: Sequence[str] = (),
) -> str:
    description = (
        f'description="{description}"' if description is not None else ""
    )
    target_role = (
        f'target_role="{target_role}"' if target_role is not None else ""
    )
    instances = "\n".join(instances)
    return f"""
        <clone
            id="{resource_id}"
            multi_state="{multi_state}"
            unique="{unique}"
            maintenance="false"
            {description}
            managed="true"
            disabled="false"
            failed="false"
            failure_ignored="false"
            {target_role}
        >
            {instances}
        </clone>
    """


def fixture_clone_dto(
    resource_id: str = "resource-clone",
    multi_state: bool = False,
    unique: bool = False,
    description: Optional[str] = None,
    target_role: Optional[str] = None,
    instances: Union[
        Sequence[PrimitiveStatusDto], Sequence[GroupStatusDto]
    ] = (),
) -> CloneStatusDto:
    return CloneStatusDto(
        resource_id,
        multi_state,
        unique,
        maintenance=False,
        description=description,
        managed=True,
        disabled=False,
        failed=False,
        failure_ignored=False,
        target_role=target_role,
        instances=list(instances),
    )


def fixture_replica_xml(
    bundle_id: str = "resource-bundle",
    replica_id: str = "0",
    bundle_type: str = "podman",
    ip: bool = False,
    node_name: str = "node1",
    member: Optional[str] = None,
) -> str:
    ip_resource = fixture_primitive_xml(
        resource_id=f"{bundle_id}-ip-192.168.122.{replica_id}",
        resource_agent="ocf:heartbeat:IPaddr2",
        node_names=[node_name],
    )
    remote_resource = fixture_primitive_xml(
        resource_id=f"{bundle_id}-{replica_id}",
        resource_agent="ocf:pacemaker:remote",
        node_names=[node_name],
    )
    container_resource = fixture_primitive_xml(
        resource_id=f"{bundle_id}-{bundle_type}-{replica_id}",
        resource_agent=f"ocf:heartbeat:{bundle_type}",
        node_names=[node_name],
    )
    return f"""
        <replica id="{replica_id}">
            {ip_resource if ip else ""}
            {member if member is not None else ""}
            {container_resource}
            {remote_resource if member is not None else ""}
        </replica>
    """


def fixture_replica_dto(
    bundle_id: str = "resource-bundle",
    replica_id: str = "0",
    bundle_type: str = "podman",
    ip: bool = False,
    node_name: str = "node1",
    member: Optional[PrimitiveStatusDto] = None,
) -> BundleReplicaStatusDto:
    ip_resource = fixture_primitive_dto(
        resource_id=f"{bundle_id}-ip-192.168.122.{replica_id}",
        resource_agent="ocf:heartbeat:IPaddr2",
        node_names=[node_name],
    )
    remote_resource = fixture_primitive_dto(
        resource_id=f"{bundle_id}-{replica_id}",
        resource_agent="ocf:pacemaker:remote",
        node_names=[node_name],
    )
    container_resource = fixture_primitive_dto(
        resource_id=f"{bundle_id}-{bundle_type}-{replica_id}",
        resource_agent=f"ocf:heartbeat:{bundle_type}",
        node_names=[node_name],
    )
    return BundleReplicaStatusDto(
        replica_id,
        member,
        remote_resource if member is not None else None,
        container_resource,
        ip_resource if ip else None,
    )


def fixture_bundle_xml(
    resource_id: str = "resource-bundle", replicas: Sequence[str] = ()
) -> str:
    replicas = "\n".join(replicas)
    return f"""
        <bundle
            id="{resource_id}"
            type="podman"
            image="localhost/pcmktest:http"
            unique="false"
            maintenance="false"
            managed="true"
            failed="false"
        >
            {replicas}
        </bundle>
    """


def fixture_bundle_dto(
    resource_id: str = "resource-bundle",
    replicas: Sequence[BundleReplicaStatusDto] = (),
) -> BundleStatusDto:
    return BundleStatusDto(
        resource_id,
        "podman",
        "localhost/pcmktest:http",
        False,
        False,
        None,
        True,
        False,
        list(replicas),
    )


def fixture_crm_mon_xml(resources: list[str]) -> str:
    # we only care about the resources element,
    # omitting other parts to make the string shorter
    resources = "\n".join(resources)
    return f"""
        <pacemaker-result
            api-version="2.30"
            request="/usr/sbin/crm_mon --one-shot --inactive --output-as xml"
        >
        <resources>
            {resources}
        </resources>
        <status code="0" message="OK"/>
        </pacemaker-result>
    """


class TestParsingErrorToReport(TestCase):
    # pylint: disable=no-self-use

    def test_empty_resource_id(self):
        report = status.cluster_status_parsing_error_to_report(
            status.EmptyResourceIdError()
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Resource with an empty id",
            ),
        )

    def test_empty_node_name(self):
        report = status.cluster_status_parsing_error_to_report(
            status.EmptyNodeNameError("resource")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Resource 'resource' contains a node with an empty name",
            ),
        )

    def test_unknown_pcmk_role(self):
        report = status.cluster_status_parsing_error_to_report(
            status.UnknownPcmkRoleError("resource", "NotPcmkRole")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Resource 'resource' contains an unknown role 'NotPcmkRole'",
            ),
        )

    def test_unexpected_member_group(self):
        report = status.cluster_status_parsing_error_to_report(
            status.UnexpectedMemberError(
                "resource", "group", "member", "bundle", ["primitive"]
            )
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason=(
                    "Unexpected resource 'member' of type 'bundle' inside of "
                    "resource 'resource' of type 'group'. Only resources of "
                    "type 'primitive' can be in a group"
                ),
            ),
        )

    def test_unexpected_member_clone(self):
        report = status.cluster_status_parsing_error_to_report(
            status.UnexpectedMemberError(
                "resource", "clone", "member", "bundle", ["primitive", "group"]
            )
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason=(
                    "Unexpected resource 'member' of type 'bundle' inside of "
                    "resource 'resource' of type 'clone'. Only resources of "
                    "type 'group', 'primitive' can be in a clone"
                ),
            ),
        )

    def test_mixed_members(self):
        report = status.cluster_status_parsing_error_to_report(
            status.MixedMembersError("resource")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Primitive and group members mixed in clone 'resource'",
            ),
        )

    def test_different_member_ids(self):
        report = status.cluster_status_parsing_error_to_report(
            status.DifferentMemberIdsError("resource")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Members with different ids in clone 'resource'",
            ),
        )

    def test_bundle_replica_missing_implicit(self):
        report = status.cluster_status_parsing_error_to_report(
            status.BundleReplicaMissingImplicitResourceError(
                "resource", "0", "container"
            )
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Replica '0' of bundle 'resource' is missing implicit container resource",
            ),
        )

    def test_bundle_replica_invalid_member_count(self):
        report = status.cluster_status_parsing_error_to_report(
            status.BundleReplicaInvalidMemberCountError("resource", "0")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Replica '0' of bundle 'resource' has invalid number of members",
            ),
        )

    def test_bundle_different_replicas(self):
        report = status.cluster_status_parsing_error_to_report(
            status.BundleDifferentReplicas("resource")
        )
        assert_report_item_equal(
            report,
            fixture.error(
                reports.codes.BAD_CLUSTER_STATE_DATA,
                reason="Replicas of bundle 'resource' are not the same",
            ),
        )


class TestPrimitiveStatusToDto(TestCase):
    # pylint: disable=protected-access
    def test_simple(self):
        primitive_xml = etree.fromstring(fixture_primitive_xml())

        result = status._primitive_to_dto(primitive_xml)

        self.assertEqual(result, fixture_primitive_dto())

    def test_empty_node_list(self):
        primitive_xml = etree.fromstring(
            fixture_primitive_xml(role=PCMK_STATUS_ROLE_STOPPED, node_names=[])
        )
        result = status._primitive_to_dto(primitive_xml)

        self.assertEqual(
            result,
            fixture_primitive_dto(role=PCMK_STATUS_ROLE_STOPPED, node_names=[]),
        )

    def test_all_attributes(self):
        primitive_xml = etree.fromstring(
            fixture_primitive_xml(
                target_role=PCMK_STATUS_ROLE_STOPPED, add_optional_args=True
            )
        )

        result = status._primitive_to_dto(primitive_xml)

        self.assertEqual(
            result,
            fixture_primitive_dto(
                target_role=PCMK_STATUS_ROLE_STOPPED, add_optional_args=True
            ),
        )

    def test_remove_clone_suffix(self):
        primitive_xml = etree.fromstring(
            fixture_primitive_xml(resource_id="resource:0")
        )

        result = status._primitive_to_dto(primitive_xml, True)

        self.assertEqual(result, fixture_primitive_dto(instance_id="0"))

    def test_running_on_multiple_nodes(self):
        primitive_xml = etree.fromstring(
            fixture_primitive_xml(node_names=["node1", "node2", "node3"])
        )

        result = status._primitive_to_dto(primitive_xml)

        self.assertEqual(
            result,
            fixture_primitive_dto(node_names=["node1", "node2", "node3"]),
        )

    def test_empty_node_name(self):
        primitive_xml = etree.fromstring(fixture_primitive_xml(node_names=[""]))

        with self.assertRaises(status.EmptyNodeNameError) as cm:
            status._primitive_to_dto(primitive_xml)
        self.assertEqual(cm.exception.resource_id, "resource")

    def test_empty_resource_id(self):
        primitive_xml = etree.fromstring(fixture_primitive_xml(resource_id=""))

        with self.assertRaises(status.EmptyResourceIdError):
            status._primitive_to_dto(primitive_xml)

    def test_role(self):
        for role in PCMK_STATUS_ROLES:
            with self.subTest(value=role):
                primitive_xml = etree.fromstring(
                    fixture_primitive_xml(role=role)
                )
                result = status._primitive_to_dto(primitive_xml)
                self.assertEqual(result, fixture_primitive_dto(role=role))

    def test_invalid_role(self):
        primitive_xml = etree.fromstring(
            fixture_primitive_xml(role="NotPcmkRole")
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._primitive_to_dto(primitive_xml)
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_target_role(self):
        for role in PCMK_ROLES:
            with self.subTest(value=role):
                primitive_xml = etree.fromstring(
                    fixture_primitive_xml(target_role=role)
                )

                result = status._primitive_to_dto(primitive_xml)

                self.assertEqual(
                    result, fixture_primitive_dto(target_role=role)
                )

    def test_invalid_target_role(self):
        for value in PCMK_STATUS_ROLES_PENDING + ("NotPcmkRole",):
            with self.subTest(value=value):
                primitive_xml = etree.fromstring(
                    fixture_primitive_xml(target_role=value)
                )

                with self.assertRaises(status.UnknownPcmkRoleError) as cm:
                    status._primitive_to_dto(primitive_xml)
                self.assertEqual(cm.exception.resource_id, "resource")
                self.assertEqual(cm.exception.role, value)

    def test_target_role_ignore_case(self):
        for value in ["started", "STARTED", "sTaRtEd"]:
            with self.subTest(value=value):
                primitive_xml = etree.fromstring(
                    fixture_primitive_xml(target_role=value)
                )
                result = status._primitive_to_dto(primitive_xml)
                self.assertEqual(
                    result, fixture_primitive_dto(target_role=PCMK_ROLE_STARTED)
                )


class TestGroupStatusToDto(TestCase):
    # pylint: disable=protected-access
    def test_all_attributes(self):
        group_xml = etree.fromstring(
            fixture_group_xml(description="Test description")
        )

        result = status._group_to_dto(group_xml)

        self.assertEqual(
            result, fixture_group_dto(description="Test description")
        )

    def test_single_member(self):
        group_xml = etree.fromstring(
            fixture_group_xml(members=[fixture_primitive_xml()])
        )

        result = status._group_to_dto(group_xml)

        self.assertEqual(
            result, fixture_group_dto(members=[fixture_primitive_dto()])
        )

    def test_multiple_members(self):
        group_xml = etree.fromstring(
            fixture_group_xml(
                members=[
                    fixture_primitive_xml(resource_id="resource1"),
                    fixture_primitive_xml(resource_id="resource2"),
                ]
            )
        )

        result = status._group_to_dto(group_xml)

        self.assertEqual(
            result,
            fixture_group_dto(
                members=[
                    fixture_primitive_dto(resource_id="resource1"),
                    fixture_primitive_dto(resource_id="resource2"),
                ]
            ),
        )

    def test_multiple_members_different_state(self):
        group_xml = etree.fromstring(
            fixture_group_xml(
                members=[
                    fixture_primitive_xml(
                        resource_id="resource1",
                        role=PCMK_STATUS_ROLE_STOPPED,
                        managed=False,
                        node_names=[],
                    ),
                    fixture_primitive_xml(resource_id="resource2"),
                ]
            )
        )

        result = status._group_to_dto(group_xml)

        self.assertEqual(
            result,
            fixture_group_dto(
                members=[
                    fixture_primitive_dto(
                        resource_id="resource1",
                        role=PCMK_STATUS_ROLE_STOPPED,
                        managed=False,
                        node_names=[],
                    ),
                    fixture_primitive_dto(resource_id="resource2"),
                ]
            ),
        )

    def test_member_invalid_role(self):
        group_xml = etree.fromstring(
            fixture_group_xml(
                members=[fixture_primitive_xml(role="NotPcmkRole")]
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._group_to_dto(group_xml)
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_member_invalid_target_role(self):
        group_xml = etree.fromstring(
            fixture_group_xml(
                members=[fixture_primitive_xml(target_role="NotPcmkRole")]
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._group_to_dto(group_xml)
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_invalid_member(self):
        resources = {
            "inner-group": '<group id="inner-group" number_resources="0" maintenance="false" managed="true" disabled="false" />',
            "inner-clone": '<clone id="inner-clone" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false" />',
            "inner-bundle": '<bundle id="inner-bundle" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false" />',
        }

        for resource_id, member in resources.items():
            with self.subTest(value=resource_id):
                group_xml = etree.fromstring(
                    fixture_group_xml(
                        resource_id="outer-group", members=[member]
                    )
                )

                with self.assertRaises(status.UnexpectedMemberError) as cm:
                    status._group_to_dto(group_xml)
                self.assertEqual(cm.exception.resource_id, "outer-group")
                self.assertEqual(cm.exception.resource_type, "group")
                self.assertEqual(cm.exception.member_id, resource_id)
                self.assertEqual(
                    cm.exception.member_type, resource_id.split("-")[1]
                )
                self.assertEqual(cm.exception.expected_types, ["primitive"])

    def test_remove_clone_suffix(self):
        group_xml = etree.fromstring(
            fixture_group_xml(
                resource_id="resource-group:0",
                members=[fixture_primitive_xml(resource_id="resource:0")],
            )
        )

        result = status._group_to_dto(group_xml, True)
        self.assertEqual(
            result,
            fixture_group_dto(
                instance_id="0",
                members=[fixture_primitive_dto(instance_id="0")],
            ),
        )


class TestCloneStatusToDto(TestCase):
    # pylint: disable=protected-access
    def test_all_attributes(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                description="Test description",
                target_role=PCMK_STATUS_ROLE_STARTED,
            )
        )

        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                description="Test description", target_role=PCMK_ROLE_STARTED
            ),
        )

    def test_primitive_member(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(instances=[fixture_primitive_xml()])
        )

        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result, fixture_clone_dto(instances=[fixture_primitive_dto()])
        )

    def test_primitive_member_multiple(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_primitive_xml(),
                    fixture_primitive_xml(node_names=["node2"]),
                ]
            )
        )

        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                instances=[
                    fixture_primitive_dto(),
                    fixture_primitive_dto(node_names=["node2"]),
                ]
            ),
        )

    def test_primitive_member_unique(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                unique=True,
                instances=[
                    fixture_primitive_xml(resource_id="resource:0"),
                    fixture_primitive_xml(
                        resource_id="resource:1", node_names=["node2"]
                    ),
                ],
            )
        )

        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                unique=True,
                instances=[
                    fixture_primitive_dto(instance_id="0"),
                    fixture_primitive_dto(
                        instance_id="1", node_names=["node2"]
                    ),
                ],
            ),
        )

    def test_primitive_member_promotable(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                multi_state=True,
                instances=[
                    fixture_primitive_xml(role=PCMK_STATUS_ROLE_PROMOTED),
                    fixture_primitive_xml(
                        role=PCMK_STATUS_ROLE_UNPROMOTED, node_names=["node2"]
                    ),
                ],
            )
        )
        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                multi_state=True,
                instances=[
                    fixture_primitive_dto(role=PCMK_STATUS_ROLE_PROMOTED),
                    fixture_primitive_dto(
                        role=PCMK_STATUS_ROLE_UNPROMOTED, node_names=["node2"]
                    ),
                ],
            ),
        )

    def test_primitive_member_invalid_role(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[fixture_primitive_xml(role="NotPcmkRole")]
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_primitive_member_invalid_target_role(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[fixture_primitive_xml(target_role="NotPcmkRole")]
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_primitive_member_different_ids(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_primitive_xml(),
                    fixture_primitive_xml(
                        resource_id="not_the_same_id", node_names=["node2"]
                    ),
                ]
            )
        )

        with self.assertRaises(status.DifferentMemberIdsError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource-clone")

    def test_group_member(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_group_xml(
                        resource_id="resource-group:0",
                        members=[fixture_primitive_xml()],
                    ),
                    fixture_group_xml(
                        resource_id="resource-group:1",
                        members=[fixture_primitive_xml(node_names=["node2"])],
                    ),
                ],
            )
        )
        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                instances=[
                    fixture_group_dto(
                        instance_id="0", members=[fixture_primitive_dto()]
                    ),
                    fixture_group_dto(
                        instance_id="1",
                        members=[fixture_primitive_dto(node_names=["node2"])],
                    ),
                ],
            ),
        )

    def test_group_member_unique(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                unique=True,
                instances=[
                    fixture_group_xml(
                        resource_id="resource-group:0",
                        members=[
                            fixture_primitive_xml(resource_id="resource:0")
                        ],
                    ),
                    fixture_group_xml(
                        resource_id="resource-group:1",
                        members=[
                            fixture_primitive_xml(
                                resource_id="resource:1", node_names=["node2"]
                            )
                        ],
                    ),
                ],
            )
        )
        result = status._clone_to_dto(clone_xml)

        self.assertEqual(
            result,
            fixture_clone_dto(
                unique=True,
                instances=[
                    fixture_group_dto(
                        instance_id="0",
                        members=[fixture_primitive_dto(instance_id="0")],
                    ),
                    fixture_group_dto(
                        instance_id="1",
                        members=[
                            fixture_primitive_dto(
                                instance_id="1", node_names=["node2"]
                            )
                        ],
                    ),
                ],
            ),
        )

    def test_group_member_different_group_ids(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_group_xml(
                        resource_id="resource-group:0",
                        members=[fixture_primitive_xml()],
                    ),
                    fixture_group_xml(
                        resource_id="another-id-:1",
                        members=[fixture_primitive_xml(node_names=["node2"])],
                    ),
                ],
            )
        )

        with self.assertRaises(status.DifferentMemberIdsError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource-clone")

    def test_group_member_different_primitive_ids(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_group_xml(
                        resource_id="resource-group:0",
                        members=[fixture_primitive_xml()],
                    ),
                    fixture_group_xml(
                        resource_id="resource-group:1",
                        members=[
                            fixture_primitive_xml(
                                resource_id="some-other-id",
                                node_names=["node2"],
                            )
                        ],
                    ),
                ],
            )
        )

        with self.assertRaises(status.DifferentMemberIdsError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource-clone")

    def test_primitive_member_types_mixed(self):
        clone_xml = etree.fromstring(
            fixture_clone_xml(
                instances=[
                    fixture_group_xml(
                        resource_id="resource",
                        members=[
                            fixture_primitive_xml(resource_id="inner-resource")
                        ],
                    ),
                    fixture_primitive_xml(node_names=["node2"]),
                ],
            )
        )

        with self.assertRaises(status.MixedMembersError) as cm:
            status._clone_to_dto(clone_xml)
        self.assertEqual(cm.exception.resource_id, "resource-clone")

    def test_invalid_member(self):
        resources = {
            "inner-clone": '<clone id="inner-clone" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false" />',
            "inner-bundle": '<bundle id="inner-bundle" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false" />',
        }
        for resource_id, element in resources.items():
            with self.subTest(value=resource_id):
                clone_xml = etree.fromstring(
                    fixture_clone_xml(
                        resource_id="outer-clone", instances=[element]
                    )
                )

                with self.assertRaises(status.UnexpectedMemberError) as cm:
                    status._clone_to_dto(clone_xml)
                self.assertEqual(cm.exception.resource_id, "outer-clone")
                self.assertEqual(cm.exception.resource_type, "clone")
                self.assertEqual(cm.exception.member_id, resource_id)
                self.assertEqual(
                    cm.exception.member_type, resource_id.split("-")[1]
                )
                self.assertEqual(
                    cm.exception.expected_types, ["primitive", "group"]
                )


class TestBundleReplicaStatusToDto(TestCase):
    # pylint: disable=protected-access
    def setUp(self):
        self.bundle_id = "resource-bundle"
        self.bundle_type = "podman"

    def test_no_member_no_ip(self):
        replica_xml = etree.fromstring(fixture_replica_xml())

        result = status._replica_to_dto(
            replica_xml, self.bundle_id, self.bundle_type
        )
        self.assertEqual(result, fixture_replica_dto())

    def test_no_member(self):
        replica_xml = etree.fromstring(fixture_replica_xml(ip=True))

        result = status._replica_to_dto(
            replica_xml, self.bundle_id, self.bundle_type
        )
        self.assertEqual(result, fixture_replica_dto(ip=True))

    def test_member(self):
        replica_xml = etree.fromstring(
            fixture_replica_xml(
                ip=True,
                member=fixture_primitive_xml(
                    node_names=["resource-bundle-0"],
                ),
            )
        )

        result = status._replica_to_dto(
            replica_xml, self.bundle_id, self.bundle_type
        )
        self.assertEqual(
            result,
            fixture_replica_dto(
                ip=True,
                member=fixture_primitive_dto(node_names=["resource-bundle-0"]),
            ),
        )

    def test_member_unique(self):
        replica_xml = etree.fromstring(
            fixture_replica_xml(
                ip=True,
                member=fixture_primitive_xml(
                    resource_id="resource:0",
                    node_names=["resource-bundle-0"],
                ),
            )
        )

        result = status._replica_to_dto(
            replica_xml, self.bundle_id, self.bundle_type, True
        )
        self.assertEqual(
            result,
            fixture_replica_dto(
                ip=True,
                member=fixture_primitive_dto(
                    instance_id="0", node_names=["resource-bundle-0"]
                ),
            ),
        )

    def test_member_no_ip(self):
        replica_xml = etree.fromstring(
            fixture_replica_xml(
                member=fixture_primitive_xml(
                    node_names=["resource-bundle-0"],
                ),
            )
        )

        result = status._replica_to_dto(
            replica_xml, self.bundle_id, self.bundle_type
        )
        self.assertEqual(
            result,
            fixture_replica_dto(
                member=fixture_primitive_dto(node_names=["resource-bundle-0"])
            ),
        )

    def test_invalid_role(self):
        replica_xml = etree.fromstring(
            fixture_replica_xml(
                member=fixture_primitive_xml(role="NotPcmkRole")
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_invalid_target_role(self):
        replica_xml = etree.fromstring(
            fixture_replica_xml(
                member=fixture_primitive_xml(target_role="NotPcmkRole")
            )
        )

        with self.assertRaises(status.UnknownPcmkRoleError) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, "resource")
        self.assertEqual(cm.exception.role, "NotPcmkRole")

    def test_no_container(self):
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
                <resource id="apa" resource_agent="ocf:heartbeat:apache" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
          </replica>
          """
        )

        with self.assertRaises(
            status.BundleReplicaMissingImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, self.bundle_id)
        self.assertEqual(cm.exception.replica_id, "0")
        self.assertEqual(cm.exception.implicit_type, "container")

    def test_empty_replica(self):
        replica_xml = etree.fromstring('<replica id="0" />')

        with self.assertRaises(
            status.BundleReplicaMissingImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, self.bundle_id)
        self.assertEqual(cm.exception.replica_id, "0")
        self.assertEqual(cm.exception.implicit_type, "container")

    def test_member_no_remote(self):
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="apa" resource_agent="ocf:heartbeat:apache" role="Stopped" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
            </replica>
            """
        )

        with self.assertRaises(
            status.BundleReplicaMissingImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, self.bundle_id)
        self.assertEqual(cm.exception.replica_id, "0")
        self.assertEqual(cm.exception.implicit_type, "remote")

    def test_member_same_id_as_container(self):
        # xml taken from crm_mon output
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="r92-1" id="1" cached="true"/>
                </resource>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="4">
                    <node name="r92-1" id="1" cached="true"/>
                    <node name="resource-bundle-1" id="resource-bundle-1" cached="true"/>
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                    <node name="resource-bundle-2" id="resource-bundle-2" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="r92-1" id="1" cached="true"/>
                </resource>
            </replica>
            """
        )
        with self.assertRaises(
            status.BundleSameIdAsImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.bundle_id, self.bundle_id)
        self.assertEqual(cm.exception.bad_ids, ["resource-bundle-podman-0"])

    def test_member_same_id_as_remote(self):
        # xml taken from crm_mon output
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Stopping" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="true" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="true" failure_ignored="false" nodes_running_on="1">
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="false"/>
                </resource>
            </replica>
            """
        )
        with self.assertRaises(
            status.BundleSameIdAsImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.bundle_id, self.bundle_id)
        self.assertEqual(cm.exception.bad_ids, ["resource-bundle-0"])

    def test_member_same_id_as_ip(self):
        # xml taken from crm_mon output
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="4">
                    <node name="node1" id="1" cached="true"/>
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                    <node name="resource-bundle-1" id="resource-bundle-1" cached="true"/>
                    <node name="resource-bundle-2" id="resource-bundle-2" cached="true"/>
                </resource>
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="node1" id="1" cached="true"/>
                </resource>
            </replica>
            """
        )
        with self.assertRaises(
            status.BundleSameIdAsImplicitResourceError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.bundle_id, self.bundle_id)
        self.assertEqual(
            cm.exception.bad_ids, ["resource-bundle-ip-192.168.122.250"]
        )

    def test_too_many_members(self):
        replica_xml = etree.fromstring(
            """
            <replica id="0">
                <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="r92-1" id="1" cached="true"/>
                </resource>
                <resource id="apa1" resource_agent="ocf:heartbeat:apache" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                </resource>
                <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="r92-1" id="1" cached="true"/>
                </resource>
                <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="r92-1" id="1" cached="true"/>
                </resource>
                <resource id="apa2" resource_agent="ocf:heartbeat:apache" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                    <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                </resource>
            </replica>
            """
        )

        with self.assertRaises(
            status.BundleReplicaInvalidMemberCountError
        ) as cm:
            status._replica_to_dto(
                replica_xml, self.bundle_id, self.bundle_type
            )
        self.assertEqual(cm.exception.resource_id, self.bundle_id)
        self.assertEqual(cm.exception.replica_id, "0")


class TestBundleStatusToDto(TestCase):
    # pylint: disable=protected-access
    def test_no_member(self):
        bundle_xml = etree.fromstring(
            fixture_bundle_xml(replicas=[fixture_replica_xml()])
        )

        result = status._bundle_to_dto(bundle_xml, False)
        self.assertEqual(
            result, fixture_bundle_dto(replicas=[fixture_replica_dto()])
        )

    def test_member(self):
        bundle_xml = etree.fromstring(
            fixture_bundle_xml(
                replicas=[
                    fixture_replica_xml(
                        ip=True,
                        member=fixture_primitive_xml(
                            node_names=["resource-bundle-0"]
                        ),
                    )
                ]
            )
        )
        result = status._bundle_to_dto(bundle_xml, False)
        self.assertEqual(
            result,
            fixture_bundle_dto(
                replicas=[
                    fixture_replica_dto(
                        ip=True,
                        member=fixture_primitive_dto(
                            node_names=["resource-bundle-0"]
                        ),
                    )
                ]
            ),
        )

    def test_multiple_replicas(self):
        bundle_xml = etree.fromstring(
            fixture_bundle_xml(
                replicas=[
                    fixture_replica_xml(
                        ip=True,
                        member=fixture_primitive_xml(
                            node_names=["resource-bundle-0"]
                        ),
                    ),
                    fixture_replica_xml(
                        ip=True,
                        replica_id="1",
                        node_name="node2",
                        member=fixture_primitive_xml(
                            node_names=["resource-bundle-1"]
                        ),
                    ),
                ]
            )
        )
        result = status._bundle_to_dto(bundle_xml, False)
        self.assertEqual(
            result,
            fixture_bundle_dto(
                replicas=[
                    fixture_replica_dto(
                        ip=True,
                        member=fixture_primitive_dto(
                            node_names=["resource-bundle-0"]
                        ),
                    ),
                    fixture_replica_dto(
                        replica_id="1",
                        ip=True,
                        node_name="node2",
                        member=fixture_primitive_dto(
                            node_names=["resource-bundle-1"]
                        ),
                    ),
                ]
            ),
        )

    def test_same_id_as_implicit(self):
        bundle_xml = etree.fromstring(
            """
            <bundle id="resource-bundle" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false">
                <replica id="0">
                    <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-0" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                    <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="2">
                        <node name="node1" id="1" cached="true"/>
                        <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                    </resource>
                </replica>
            </bundle>
            """
        )

        with self.assertRaises(
            status.BundleSameIdAsImplicitResourceError
        ) as cm:
            status._bundle_to_dto(bundle_xml, False)
        self.assertEqual(cm.exception.bundle_id, "resource-bundle")
        self.assertEqual(cm.exception.bad_ids, ["resource-bundle-0"])

    def test_same_id_as_implicit_multiple_replicas(self):
        bundle_xml = etree.fromstring(
            """
            <bundle id="resource-bundle" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false">
                <replica id="0">
                    <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-1" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                    <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                </replica>
                <replica id="1">
                    <resource id="resource-bundle-ip-192.168.122.251" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node2" id="2" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-1" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                    <resource id="resource-bundle-podman-1" resource_agent="ocf:heartbeat:podman" role="Stopping" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="true" failure_ignored="false" nodes_running_on="1">
                        <node name="node2" id="2" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-1" resource_agent="ocf:pacemaker:remote" role="Started" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="true" failure_ignored="false" nodes_running_on="1">
                        <node name="resource-bundle-1" id="resource-bundle-1" cached="false"/>
                    </resource>
                </replica>
            </bundle>
            """
        )
        with self.assertRaises(
            status.BundleSameIdAsImplicitResourceError
        ) as cm:
            status._bundle_to_dto(bundle_xml, False)
        self.assertEqual(cm.exception.bundle_id, "resource-bundle")
        self.assertEqual(cm.exception.bad_ids, ["resource-bundle-1"])

    def test_replicas_different(self):
        replicas = {
            "no-ip": fixture_replica_xml(
                ip=False, member=fixture_primitive_xml()
            ),
            "different-member-id": fixture_replica_xml(
                ip=True, member=fixture_primitive_xml(resource_id="another-id")
            ),
            "no-member": fixture_replica_xml(ip=True, member=None),
            "different-member-agent": fixture_replica_xml(
                ip=True,
                member=fixture_primitive_xml(
                    resource_agent="ocf:heartbeat:apache"
                ),
            ),
        }
        for name, element in replicas.items():
            with self.subTest(value=name):
                bundle_xml = etree.fromstring(
                    fixture_bundle_xml(
                        replicas=[
                            element,
                            fixture_replica_xml(
                                ip=True,
                                replica_id="1",
                                member=fixture_primitive_xml(),
                            ),
                        ]
                    )
                )

                with self.assertRaises(status.BundleDifferentReplicas) as cm:
                    status._bundle_to_dto(bundle_xml)
                self.assertEqual(cm.exception.resource_id, "resource-bundle")


class TestResourcesStatusToDto(TestCase):
    def test_empty_resources(self):
        status_xml = etree.fromstring(fixture_crm_mon_xml([]))

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()
        self.assertEqual(result, ResourcesStatusDto([]))
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_single_primitive(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml([fixture_primitive_xml()])
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()
        self.assertEqual(result, ResourcesStatusDto([fixture_primitive_dto()]))
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_single_group(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [fixture_group_xml(members=[fixture_primitive_xml()])]
            )
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()
        self.assertEqual(
            result,
            ResourcesStatusDto(
                [fixture_group_dto(members=[fixture_primitive_dto()])]
            ),
        )
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_single_clone(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [fixture_clone_xml(instances=[fixture_primitive_xml()])]
            )
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()
        self.assertEqual(
            result,
            ResourcesStatusDto(
                [fixture_clone_dto(instances=[fixture_primitive_dto()])]
            ),
        )
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_single_bundle(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [
                    fixture_bundle_xml(
                        replicas=[
                            fixture_replica_xml(
                                ip=True,
                                member=fixture_primitive_xml(
                                    node_names=["resource-bundle-0"]
                                ),
                            )
                        ]
                    )
                ]
            )
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()
        self.assertEqual(
            result,
            ResourcesStatusDto(
                [
                    fixture_bundle_dto(
                        replicas=[
                            fixture_replica_dto(
                                ip=True,
                                member=fixture_primitive_dto(
                                    node_names=["resource-bundle-0"]
                                ),
                            )
                        ]
                    )
                ]
            ),
        )
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_all_resource_types(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [
                    fixture_primitive_xml(),
                    fixture_group_xml(members=[fixture_primitive_xml()]),
                    fixture_clone_xml(instances=[fixture_primitive_xml()]),
                    fixture_bundle_xml(
                        replicas=[
                            fixture_replica_xml(
                                ip=True,
                                member=fixture_primitive_xml(
                                    node_names=["resource-bundle-0"]
                                ),
                            )
                        ]
                    ),
                ]
            )
        )
        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()

        self.assertEqual(
            result,
            ResourcesStatusDto(
                [
                    fixture_primitive_dto(),
                    fixture_group_dto(members=[fixture_primitive_dto()]),
                    fixture_clone_dto(instances=[fixture_primitive_dto()]),
                    fixture_bundle_dto(
                        replicas=[
                            fixture_replica_dto(
                                ip=True,
                                member=fixture_primitive_dto(
                                    node_names=["resource-bundle-0"]
                                ),
                            )
                        ]
                    ),
                ]
            ),
        )
        assert_report_item_list_equal(parser.get_warnings(), [])

    def test_skip_bundle_same_id(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [
                    fixture_primitive_xml(),
                    """
            <bundle id="resource-bundle" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false">
                <replica id="0">
                    <resource id="resource-bundle-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-0" resource_agent="ocf:heartbeat:apache" role="Stopped" active="false" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="0"/>
                    <resource id="resource-bundle-podman-0" resource_agent="ocf:heartbeat:podman" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
                        <node name="node1" id="1" cached="true"/>
                    </resource>
                    <resource id="resource-bundle-0" resource_agent="ocf:pacemaker:remote" role="Started" active="true" orphaned="false" blocked="false" maintenance="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="2">
                        <node name="node1" id="1" cached="true"/>
                        <node name="resource-bundle-0" id="resource-bundle-0" cached="true"/>
                    </resource>
                </replica>
            </bundle>
                    """,
                ]
            )
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()

        self.assertEqual(result, ResourcesStatusDto([fixture_primitive_dto()]))
        assert_report_item_list_equal(
            parser.get_warnings(),
            [
                fixture.warn(
                    reports.codes.CLUSTER_STATUS_BUNDLE_MEMBER_ID_AS_IMPLICIT,
                    bundle_id="resource-bundle",
                    bad_ids=["resource-bundle-0"],
                )
            ],
        )

    def test_skip_bundle_missing_implicit(self):
        status_xml = etree.fromstring(
            fixture_crm_mon_xml(
                [fixture_bundle_xml("bundle", ['<replica id="0"/>'])]
            )
        )

        parser = status.ClusterStatusParser(status_xml)
        result = parser.status_xml_to_dto()

        self.assertEqual(result, ResourcesStatusDto([]))
        assert_report_item_list_equal(
            parser.get_warnings(),
            [
                fixture.warn(
                    reports.codes.BAD_CLUSTER_STATE_DATA,
                    reason=(
                        "Replica '0' of bundle 'bundle' is missing implicit "
                        "container resource"
                    ),
                )
            ],
        )



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.const import PcmkRoleType
from pcs.common.interface.dto import DataTransferObject

from ..rule import CibRuleExpressionDto
from .set import CibResourceSetDto


@dataclass(frozen=True)
class CibConstraintColocationAttributesDto(DataTransferObject):
    constraint_id: str
    score: Optional[str]
    influence: Optional[str]
    lifetime: Sequence[CibRuleExpressionDto]


@dataclass(frozen=True)
class CibConstraintColocationDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    resource_id: str
    with_resource_id: str
    node_attribute: Optional[str]
    resource_role: Optional[PcmkRoleType]
    with_resource_role: Optional[PcmkRoleType]
    resource_instance: Optional[int]
    with_resource_instance: Optional[int]
    attributes: CibConstraintColocationAttributesDto


@dataclass(frozen=True)
class CibConstraintColocationSetDto(DataTransferObject):
    resource_sets: Sequence[CibResourceSetDto]
    attributes: CibConstraintColocationAttributesDto



import json
from collections import Counter
from typing import (
    Any,
    Generic,
    Mapping,
    NamedTuple,
    Optional,
    TypeVar,
    overload,
)

from pcs.common import reports
from pcs.common.str_tools import format_list

ALL_RESOURCE_XML_TAGS = ["bundle", "clone", "group", "master", "primitive"]


# Previously, a report item fixture was a plain tuple. That was, however, not
# the best to work with when access to specific indexes was needed. To provide
# a 1-to-1 replacement with more friendly interface, a NamedTuple was chosen
# instead of a dataclass. Order of the attributes is the same as in the
# original tuple, so it may be not the best, but it works without any changes
# to related code.
class ReportItemFixture(NamedTuple):
    severity: reports.types.SeverityLevel
    code: reports.types.MessageCode
    payload: Mapping[str, Any]
    force_code: Optional[reports.types.ForceCode]
    context: Optional[Mapping[str, Any]]

    def to_warn(self):
        return warn(self.code, self.context, **self.payload)

    def adapt(self, **payload):
        updated_payload = self.payload.copy()
        updated_payload.update(**payload)
        return type(self)(
            self.severity,
            self.code,
            updated_payload,
            self.force_code,
            self.context,
        )


def debug(
    code: reports.types.MessageCode,
    context: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> ReportItemFixture:
    return ReportItemFixture(
        reports.ReportItemSeverity.DEBUG, code, kwargs, None, context
    )


def warn(
    code: reports.types.MessageCode,
    context: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> ReportItemFixture:
    return ReportItemFixture(
        reports.ReportItemSeverity.WARNING, code, kwargs, None, context
    )


def deprecation(
    code: reports.types.MessageCode,
    context: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> ReportItemFixture:
    return ReportItemFixture(
        reports.ReportItemSeverity.DEPRECATION, code, kwargs, None, context
    )


def error(
    code: reports.types.MessageCode,
    force_code: Optional[reports.types.ForceCode] = None,
    context: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> ReportItemFixture:
    return ReportItemFixture(
        reports.ReportItemSeverity.ERROR, code, kwargs, force_code, context
    )


def info(
    code: reports.types.MessageCode,
    context: Optional[Mapping[str, Any]] = None,
    **kwargs,
) -> ReportItemFixture:
    return ReportItemFixture(
        reports.ReportItemSeverity.INFO, code, kwargs, None, context
    )


T = TypeVar("T")


class NameValueSequence(Generic[T]):
    def __init__(
        self,
        name_list: Optional[list[Optional[str]]] = None,
        value_list: Optional[list[T]] = None,
    ):
        self.__names: list[Optional[str]] = name_list or []
        self.__values: list[T] = value_list or []

        if len(self.__names) != len(self.__values):
            raise AssertionError("Values count doesn't match names count")

        name_counter = Counter(self.__names)
        duplicate_names = [
            n for n in name_counter if n is not None and name_counter[n] > 1
        ]
        if duplicate_names:
            raise AssertionError(
                f"Duplicate names are not allowed in {type(self).__name__}. "
                f"Found duplications:\n  {format_list(duplicate_names)}"
            )

    @property
    def names(self) -> list[Optional[str]]:
        return list(self.__names)

    @property
    def values(self) -> list[T]:
        return list(self.__values)

    def append(self, value: T, name: Optional[str] = None) -> None:
        """
        Append new value with the specified name at the end of sequence

        value -- new value
        name -- new value name
        """
        self.__check_name(name)
        self.__names.append(name)
        self.__values.append(value)

    def prepend(self, value: T, name: Optional[str] = None) -> None:
        """
        Insert new value with the specified name at the start of sequence

        value -- new value
        name -- new value name
        """
        self.__check_name(name)
        self.__names.insert(0, name)
        self.__values.insert(0, value)

    def insert(self, before: str, value: T, name: Optional[str] = None) -> None:
        """
        Insert new value before a specified value

        before -- name of a value before which the new one will be placed
        value -- the new value
        name -- name of the new value
        """
        self.__check_name(name)
        index = self.__get_index(before)
        self.__names.insert(index, name)
        self.__values.insert(index, value)

    def remove(self, *name_list: str) -> None:
        """
        Remove values with specified names

        name_list -- names of values to be removed
        """
        for name in name_list:
            index = self.__get_index(name)
            del self.__names[index]
            del self.__values[index]

    def replace(
        self, name: str, value: T, new_name: Optional[str] = None
    ) -> None:
        """
        Replace a value specified by its name

        name -- name of a value to be replaced
        value -- new value
        new_name -- new name of the value, use 'name' if not specified
        """
        if new_name is not None and new_name != name and name in self.__names:
            raise AssertionError(f"Name '{new_name}' already present in {self}")
        for i, current_name in enumerate(self.__names):
            if current_name == name:
                self.__values[i] = value
                if new_name:
                    self.__names[i] = new_name
                return
        raise IndexError(self.__index_error(name))

    def trim_before(self, name: str) -> None:
        """
        Remove a value with the specified name and all values after it

        name -- name of a value to trim at
        """
        index = self.__get_index(name)
        self.__names = self.__names[:index]
        self.__values = self.__values[:index]

    def copy(self) -> "NameValueSequence":
        return type(self)(self.names, self.values)

    def __get_index(self, name: str) -> int:
        try:
            return self.__names.index(name)
        except ValueError as e:
            raise IndexError(self.__index_error(name)) from e

    def __index_error(self, index: str) -> str:
        return f"'{index}' not present in {self}"

    def __check_name(self, name: Optional[str]) -> None:
        if name is not None and name in self.__names:
            raise AssertionError(f"Name '{name}' already present in {self}")

    @overload
    def __getitem__(self, spec: str) -> T:
        pass

    @overload
    def __getitem__(self, spec: slice) -> "NameValueSequence":
        pass

    def __getitem__(self, spec):
        if not isinstance(spec, slice):
            try:
                return self.__values[self.__names.index(spec)]
            except ValueError as e:
                raise IndexError(self.__index_error(spec)) from e

        assert spec.step is None, "Step is not supported in slicing"
        start = None if spec.start is None else self.__names.index(spec.start)
        stop = None if spec.stop is None else self.__names.index(spec.stop)
        return type(self)(self.__names[start:stop], self.__values[start:stop])

    def __setitem__(self, name: str, value: T) -> None:
        return (
            self.replace(name, value)
            if name in self.__names
            else self.append(value, name)
        )

    def __delitem__(self, name: str) -> None:
        index = self.__get_index(name)
        del self.__names[index]
        del self.__values[index]

    def __add__(self, other: "NameValueSequence") -> "NameValueSequence":
        my_name = type(self).__name__
        other_name = type(other).__name__
        assert isinstance(other, type(self)), (
            f"Can only concatenate {my_name} with {my_name}, not {other_name}"
        )

        return type(self)(
            self.names + other.names,
            self.values + other.values,
        )

    def __str__(self) -> str:
        return f"{type(self).__name__} {hex(id(self))}:\n" + "\n".join(
            [
                f" {index:3}. {item[0] if item[0] else '<unnamed>'}: {item[1]}"
                for index, item in enumerate(
                    zip(self.__names, self.__values, strict=False), 1
                )
            ]
        )


class ReportSequenceBuilder:
    def __init__(self, store: Optional[NameValueSequence] = None):
        self._store = store or NameValueSequence()

    @property
    def fixtures(self) -> NameValueSequence:
        return self._store

    def info(
        self,
        code: reports.types.MessageCode,
        _name: Optional[str] = None,
        **kwargs,
    ) -> "ReportSequenceBuilder":
        self._store.append(info(code, **kwargs), _name)
        return self

    def warn(
        self,
        code: reports.types.MessageCode,
        _name: Optional[str] = None,
        **kwargs,
    ) -> "ReportSequenceBuilder":
        self._store.append(warn(code, **kwargs), _name)
        return self

    def deprecation(
        self,
        code: reports.types.MessageCode,
        _name: Optional[str] = None,
        **kwargs,
    ) -> "ReportSequenceBuilder":
        self._store.append(deprecation(code, **kwargs), _name)
        return self

    def error(
        self,
        code: reports.types.MessageCode,
        force_code: Optional[reports.types.ForceCode] = None,
        _name: Optional[str] = None,
        **kwargs,
    ) -> "ReportSequenceBuilder":
        self._store.append(error(code, force_code=force_code, **kwargs), _name)
        return self


def report_not_found(
    res_id, context_type="", expected_types=None, context_id=""
):
    return error(
        reports.codes.ID_NOT_FOUND,
        context_type=context_type,
        context_id=context_id,
        id=res_id,
        expected_types=(
            ALL_RESOURCE_XML_TAGS if expected_types is None else expected_types
        ),
    )


def report_not_resource_or_tag(element_id, context_type="cib", context_id=""):
    return report_not_found(
        element_id,
        context_type=context_type,
        expected_types=sorted(ALL_RESOURCE_XML_TAGS + ["tag"]),
        context_id=context_id,
    )


def report_invalid_id(_id, invalid_char, id_description="id"):
    return error(
        reports.codes.INVALID_ID_BAD_CHAR,
        id=_id,
        id_description=id_description,
        is_first_char=(_id.index(invalid_char) == 0),
        invalid_character=invalid_char,
    )


def report_id_already_exist(_id):
    return error(
        reports.codes.ID_ALREADY_EXISTS,
        id=_id,
    )


def report_resource_not_running(
    resource, severity=reports.ReportItemSeverity.INFO
):
    return ReportItemFixture(
        severity,
        reports.codes.RESOURCE_DOES_NOT_RUN,
        dict(resource_id=resource),
        None,
        None,
    )


def report_resource_running(
    resource, roles, severity=reports.ReportItemSeverity.INFO
):
    return ReportItemFixture(
        severity,
        reports.codes.RESOURCE_RUNNING_ON_NODES,
        dict(
            resource_id=resource,
            roles_with_nodes=roles,
        ),
        None,
        None,
    )


def report_unexpected_element(element_id, element_type, expected_types):
    return error(
        reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
        id=element_id,
        expected_types=expected_types,
        current_type=element_type,
    )


def report_not_for_bundles(element_id):
    return report_unexpected_element(
        element_id, "bundle", ["clone", "master", "group", "primitive"]
    )


def report_wait_for_idle_timed_out(reason):
    return error(reports.codes.WAIT_FOR_IDLE_TIMED_OUT, reason=reason.strip())


def check_sbd_comm_success_fixture(node, watchdog, device_list):
    return dict(
        label=node,
        output=json.dumps(
            {
                "sbd": {
                    "installed": True,
                },
                "watchdog": {
                    "exist": True,
                    "path": watchdog,
                    "is_supported": True,
                },
                "device_list": [
                    dict(path=dev, exist=True, block_device=True)
                    for dev in device_list
                ],
            }
        ),
        param_list=[
            ("watchdog", watchdog),
            ("device_list", json.dumps(device_list)),
        ],
    )



import os
import re
from functools import partial
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    compare_version,
    get_tmp_dir,
    get_tmp_file,
    outdent,
    read_test_resource,
    skip_unless_root,
    write_file_to_tmpfile,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.pcs_runner import (
    PcsRunner,
    pcs,
)
from pcs_test.tools.xml import str_to_etree


class UidGidTest(TestCase):
    def setUp(self):
        self.uid_gid_dir = get_tmp_dir("tier1_cluster_uidgid")
        self._pcs = partial(
            pcs,
            None,
            mock_settings={"corosync_uidgid_dir": self.uid_gid_dir.name},
        )

    def tearDown(self):
        self.uid_gid_dir.cleanup()

    def assert_uidgid_file_content(self, filename, content):
        self.assertEqual(
            read_test_resource(os.path.join(self.uid_gid_dir.name, filename)),
            content,
        )

    def assert_uidgid_file_removed(self, filename):
        self.assertEqual(
            os.path.isfile(os.path.join(self.uid_gid_dir.name, filename)),
            False,
        )

    def test_uidgid(self):  # noqa: PLR0915
        # pylint: disable=too-many-statements
        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "No uidgids configured\n")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs("cluster uidgid add".split())
        self.assertEqual(stdout, "")
        self.assertTrue(stderr.startswith("\nUsage:"))
        self.assertEqual(retval, 1)

        stdout, stderr, retval = self._pcs("cluster uidgid xx".split())
        self.assertEqual(stdout, "")
        self.assertTrue(stderr.startswith("\nUsage:"))
        self.assertEqual(retval, 1)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid add uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_content(
            "pcs-uidgid-testuid-testgid",
            outdent(
                """\
                uidgid {
                  uid: testuid
                  gid: testgid
                }
                """
            ),
        )

        stdout, stderr, retval = self._pcs(
            "cluster uidgid add uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Error: uidgid file with uid=testuid and gid=testgid already "
            "exists\n",
        )
        self.assertEqual(retval, 1)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid delete uid=testuid2 gid=testgid2".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Error: no uidgid files with uid=testuid2 and gid=testgid2 found\n",
        )
        self.assertEqual(retval, 1)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid remove uid=testuid gid=testgid2".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Error: no uidgid files with uid=testuid and gid=testgid2 found\n",
        )
        self.assertEqual(retval, 1)

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "UID/GID: uid=testuid gid=testgid\n")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid delete uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_removed("pcs-uidgid-testuid-testgid")

        stdout, stderr, retval = self._pcs(
            "cluster uidgid add uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "UID/GID: uid=testuid gid=testgid\n")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid remove uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid add uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_content(
            "pcs-uidgid-testuid-testgid",
            outdent(
                """\
                uidgid {
                  uid: testuid
                  gid: testgid
                }
                """
            ),
        )

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "UID/GID: uid=testuid gid=testgid\n")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid delete uid=testuid gid=testgid".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_removed("pcs-uidgid-testuid-testgid")

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "No uidgids configured\n")
        self.assertEqual(retval, 0)

    def test_missing_uid_gid(self):
        stdout, stderr, retval = self._pcs(
            "cluster uidgid add uid=1000".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_content(
            "pcs-uidgid-1000-",
            outdent(
                """\
                uidgid {
                  uid: 1000
                }
                """
            ),
        )

        stdout, stderr, retval = self._pcs(
            "cluster uidgid add gid=1001".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_content(
            "pcs-uidgid--1001",
            outdent(
                """\
                uidgid {
                  gid: 1001
                }
                """
            ),
        )

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(
            stdout,
            outdent(
                """\
                UID/GID: uid= gid=1001
                UID/GID: uid=1000 gid=
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        stdout, stderr, retval = self._pcs(
            "cluster uidgid delete uid=1000".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_removed("pcs-uidgid-1000-")

        stdout, stderr, retval = self._pcs(
            "cluster uidgid delete gid=1001".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        self.assert_uidgid_file_removed("pcs-uidgid--1001")

        stdout, stderr, retval = self._pcs("cluster uidgid".split())
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "No uidgids configured\n")
        self.assertEqual(retval, 0)


class ClusterUpgradeTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cluster_upgrade")
        write_file_to_tmpfile(rc("cib-empty-1.2.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_cluster_upgrade(self):
        def extract_version(string_value):
            match = re.match(
                r"^pacemaker-(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<rev>\d+))?$",
                string_value,
            )
            return (
                int(match.group("major")),
                int(match.group("minor")),
                int(match.group("rev")) if match.group("rev") else 0,
            )

        self.temp_cib.seek(0)
        cib = str_to_etree(self.temp_cib.read())
        validate_with = cib.getroottree().getroot().get("validate-with")
        self.assertEqual(validate_with, "pacemaker-1.2")
        version_before = extract_version(validate_with)
        self.assertEqual(version_before, (1, 2, 0))

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "cluster cib-upgrade".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr, "Cluster CIB has been upgraded to latest version\n"
        )
        self.assertEqual(retval, 0)

        self.temp_cib.seek(0)
        cib = str_to_etree(self.temp_cib.read())
        validate_with = cib.getroottree().getroot().get("validate-with")
        version_after = extract_version(validate_with)
        self.assertEqual(compare_version(version_before, version_after), -1)

        stdout, stderr, retval = pcs(
            self.temp_cib.name, "cluster cib-upgrade".split()
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr, "Cluster CIB has been upgraded to latest version\n"
        )
        self.assertEqual(retval, 0)

        self.temp_cib.seek(0)
        cib = str_to_etree(self.temp_cib.read())
        validate_with = cib.getroottree().getroot().get("validate-with")
        version_after2 = extract_version(validate_with)
        self.assertEqual(compare_version(version_after, version_after2), 0)


@skip_unless_root()
class ClusterStartStop(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(cib_file=None)

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "cluster stop rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "cluster start rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )


@skip_unless_root()
class ClusterEnableDisable(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(cib_file=None)

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "cluster enable rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "cluster disable rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )



from unittest import TestCase

from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)
from pcs.lib.cib.rule import tools
from pcs.lib.cib.rule.expression_part import (
    BOOL_AND,
    BOOL_OR,
    NODE_ATTR_OP_EQ,
    BoolExpr,
    NodeAttrExpr,
    OpExpr,
    RscExpr,
)


class IsRscExpressionsOnly(TestCase):
    def test_rsc_expr(self):
        self.assertTrue(
            tools.is_rsc_expressions_only(RscExpr("ocf", None, None))
        )

    def test_rsc_expr_in_tree(self):
        self.assertTrue(
            tools.is_rsc_expressions_only(
                BoolExpr(
                    BOOL_OR,
                    [
                        BoolExpr(
                            BOOL_AND,
                            [
                                RscExpr("ocf", None, None),
                                RscExpr(None, None, "Dummy"),
                            ],
                        ),
                        RscExpr(None, "pacemaker", None),
                    ],
                )
            )
        )

    def test_no_rsc_expr(self):
        self.assertFalse(tools.is_rsc_expressions_only(OpExpr("monitor", None)))

    def test_no_rsc_expr_in_tree(self):
        self.assertFalse(
            tools.is_rsc_expressions_only(
                BoolExpr(
                    BOOL_AND,
                    [
                        OpExpr("monitor", None),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "foo", "bar", None),
                    ],
                ),
            )
        )

    def test_mixed(self):
        self.assertFalse(
            tools.is_rsc_expressions_only(
                BoolExpr(
                    BOOL_OR,
                    [
                        BoolExpr(
                            BOOL_AND,
                            [
                                RscExpr("ocf", None, None),
                                NodeAttrExpr(
                                    NODE_ATTR_OP_EQ, "foo", "bar", None
                                ),
                            ],
                        ),
                        RscExpr(None, "pacemaker", None),
                    ],
                )
            )
        )


class IsRscExpressionsOnlyDto(TestCase):
    dto_id = 0

    def fixture_dto(self, dto_type, children):
        self.dto_id += 1
        # dto is not complete and valid, it only contains parts which matter
        # for the tested functionality
        return CibRuleExpressionDto(
            f"id{self.dto_id}",
            dto_type,
            CibRuleInEffectStatus.UNKNOWN,
            dict(),
            None,
            None,
            children,
            "",
        )

    def test_rsc_expr(self):
        self.assertTrue(
            tools.is_rsc_expressions_only_dto(
                self.fixture_dto(CibRuleExpressionType.RSC_EXPRESSION, [])
            )
        )

    def test_rsc_expr_in_tree(self):
        self.assertTrue(
            tools.is_rsc_expressions_only_dto(
                self.fixture_dto(
                    CibRuleExpressionType.RULE,
                    [
                        self.fixture_dto(
                            CibRuleExpressionType.RULE,
                            [
                                self.fixture_dto(
                                    CibRuleExpressionType.RSC_EXPRESSION, []
                                ),
                                self.fixture_dto(
                                    CibRuleExpressionType.RSC_EXPRESSION, []
                                ),
                            ],
                        ),
                        self.fixture_dto(
                            CibRuleExpressionType.RSC_EXPRESSION, []
                        ),
                    ],
                )
            )
        )

    def test_no_rsc_expr(self):
        self.assertFalse(
            tools.is_rsc_expressions_only_dto(
                self.fixture_dto(CibRuleExpressionType.OP_EXPRESSION, [])
            )
        )

    def test_no_rsc_expr_in_tree(self):
        self.assertFalse(
            tools.is_rsc_expressions_only_dto(
                self.fixture_dto(
                    CibRuleExpressionType.RULE,
                    [
                        self.fixture_dto(
                            CibRuleExpressionType.OP_EXPRESSION, []
                        ),
                        self.fixture_dto(CibRuleExpressionType.EXPRESSION, []),
                    ],
                )
            )
        )

    def test_mixed(self):
        self.assertFalse(
            tools.is_rsc_expressions_only_dto(
                self.fixture_dto(
                    CibRuleExpressionType.RULE,
                    [
                        self.fixture_dto(
                            CibRuleExpressionType.RULE,
                            [
                                self.fixture_dto(
                                    CibRuleExpressionType.RSC_EXPRESSION, []
                                ),
                                self.fixture_dto(
                                    CibRuleExpressionType.EXPRESSION, []
                                ),
                            ],
                        ),
                        self.fixture_dto(
                            CibRuleExpressionType.RSC_EXPRESSION, []
                        ),
                    ],
                )
            )
        )



from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.fixture_cib import modify_cib_file
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


def fixture_stonith_primitive_xml(resource_id: str) -> str:
    return f"""
        <primitive id="{resource_id}" class="stonith" type="fence_pcsmock_minimal"/>
    """


FIXTURE_TAG_XML = """
    <tag id="TAG">
        <obj_ref id="S2"/>
        <obj_ref id="S3"/>
    </tag>
"""
FIXTURE_SET_CONSTRAINT = """
    <rsc_colocation score="10" id="colocation-constraint">
        <resource_set role="Started" id="set1">
            <resource_ref id="S2"/>
            <resource_ref id="S3"/>
        </resource_set>
    </rsc_colocation>
"""
FIXTURE_FENCING_LEVEL_XML = """
    <fencing-level index="3" devices="S2,S3" target="foo" id="fencing-level"/>
"""

ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)
NO_STONITH_LEFT_ERROR = (
    "Error: Requested action leaves the cluster with no enabled means to fence "
    "nodes, resulting in the cluster not being able to recover from certain "
    "failure conditions, use --force to override\n"
)
NO_STONITH_LEFT_WARNING = (
    "Warning: Requested action leaves the cluster with no enabled means to fence "
    "nodes, resulting in the cluster not being able to recover from certain "
    "failure conditions\n"
)


class StonithRemoveDeleteBase(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).find(".//resources"))
    ),
):
    command = ""

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_stonith_remove")
        cib_data = modify_cib_file(
            get_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_stonith_primitive_xml("S1")}
                    {fixture_stonith_primitive_xml("S2")}
                    {fixture_stonith_primitive_xml("S3")}
                </resources>
            """,
            constraints=f"""
                <constraints>
                    {FIXTURE_SET_CONSTRAINT}
                </constraints>
            """,
            tags=f"""
                <tags>
                    {FIXTURE_TAG_XML}
                </tags>
            """,
            fencing_topology=f"""
                <fencing-topology>
                    {FIXTURE_FENCING_LEVEL_XML}
                </fencing-topology>
            """,
        )
        write_data_to_tmpfile(cib_data, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def assert_constraints(self, constraints_xml):
        self.assert_resources_xml_in_cib(
            constraints_xml,
            lambda cib: etree.tostring(etree.parse(cib).find(".//constraints")),
        )

    def assert_tags(self, tags_xml):
        self.assert_resources_xml_in_cib(
            tags_xml,
            lambda cib: etree.tostring(etree.parse(cib).find(".//tags")),
        )

    def assert_fencing_topology(self, fencing_xml):
        self.assert_resources_xml_in_cib(
            fencing_xml,
            lambda cib: etree.tostring(
                etree.parse(cib).find(".//fencing-topology")
            ),
        )

    def tearDown(self):
        self.temp_cib.close()

    def test_no_args(self):
        self.assert_pcs_fail(
            ["stonith", self.command],
            stderr_start=f"\nUsage: pcs stonith {self.command}...",
        )

    def test_nonexistent_resource(self):
        self.assert_pcs_fail(
            ["stonith", self.command, "nonexistent"],
            stderr_full="Error: Unable to find stonith resource: 'nonexistent'\n",
        )

    def test_single_resource(self):
        self.assert_effect_single(
            ["stonith", self.command, "S1"],
            f"""
            <resources>
                {fixture_stonith_primitive_xml("S2")}
                {fixture_stonith_primitive_xml("S3")}
            </resources>
            """,
            stderr_full="",
        )
        self.assert_constraints(
            f"""
            <constraints>
                {FIXTURE_SET_CONSTRAINT}
            </constraints>
            """
        )
        self.assert_tags(
            f"""
            <tags>
                {FIXTURE_TAG_XML}
            </tags>
            """
        )
        self.assert_fencing_topology(
            f"""
            <fencing-topology>
                {FIXTURE_FENCING_LEVEL_XML}
            </fencing-topology>
            """
        )

    def test_remove_references(self):
        self.assert_effect_single(
            ["stonith", self.command, "S2"],
            f"""
            <resources>
                {fixture_stonith_primitive_xml("S1")}
                {fixture_stonith_primitive_xml("S3")}
            </resources>
            """,
            stderr_full=dedent(
                """\
                Removing references:
                  Resource 'S2' from:
                    Fencing level: 'fencing-level'
                    Resource set: 'set1'
                    Tag: 'TAG'
                """
            ),
        )
        self.assert_constraints(
            """
            <constraints>
                <rsc_colocation score="10" id="colocation-constraint">
                    <resource_set role="Started" id="set1">
                        <resource_ref id="S3"/>
                    </resource_set>
                </rsc_colocation>
            </constraints>
            """
        )
        self.assert_tags(
            """
            <tags>
                <tag id="TAG">
                    <obj_ref id="S3"/>
                </tag>
            </tags>
            """
        )
        self.assert_fencing_topology(
            """
            <fencing-topology>
                <fencing-level index="3" devices="S3" target="foo" id="fencing-level"/>
            </fencing-topology>
            """
        )

    def test_remove_all_resources(self):
        self.assert_pcs_fail(
            ["stonith", self.command, "S1", "S2", "S3"],
            NO_STONITH_LEFT_ERROR + ERRORS_HAVE_OCCURRED,
        )
        self.assert_effect_single(
            ["stonith", self.command, "S1", "S2", "S3", "--force"],
            "<resources/>",
            stderr_full=(
                NO_STONITH_LEFT_WARNING
                + dedent(
                    """\
                    Removing dependant elements:
                      Colocation constraint: 'colocation-constraint'
                      Fencing level: 'fencing-level'
                      Resource set: 'set1'
                      Tag: 'TAG'
                    """
                )
            ),
        )
        self.assert_constraints("<constraints/>")
        self.assert_tags("<tags/>")
        self.assert_fencing_topology("<fencing-topology/>")


class StonithRemove(StonithRemoveDeleteBase, TestCase):
    command = "remove"


class StonithDelete(StonithRemoveDeleteBase, TestCase):
    command = "delete"


class StonithReferencedInAcl(AssertPcsMixin, TestCase):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_stonith_remove_referenced_in_acl")
        cib_data = modify_cib_file(
            get_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_stonith_primitive_xml("S1")}
                </resources>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="ROLE">
                        <acl_permission id="PERMISSION" kind="write" reference="S1"/>
                    </acl_role>
                    <acl_target id="TARGET">
                        <role id="ROLE"/>
                    </acl_target>
                </acls>
            """,
        )
        write_data_to_tmpfile(cib_data, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_remove_primitive(self):
        self.assert_pcs_fail(
            ["stonith", "delete", "S1"],
            NO_STONITH_LEFT_ERROR + ERRORS_HAVE_OCCURRED,
        )
        self.assert_pcs_success(
            ["stonith", "delete", "S1", "--force"],
            stderr_full=(
                NO_STONITH_LEFT_WARNING
                + dedent(
                    """\
                    Removing dependant element:
                      Acl permission: 'PERMISSION'
                    Removing references:
                      Acl permission 'PERMISSION' from:
                        Acl role: 'ROLE'
                    """
                )
            ),
        )



import logging
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common import pcs_pycurl as pycurl
from pcs.common.communication import logger
from pcs.common.host import Destination
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
    Response,
)
from pcs.common.reports import codes as report_codes
from pcs.common.reports.processor import ReportProcessorToLog

from pcs_test.tools import fixture
from pcs_test.tools.custom_mock import (
    MockCurlSimple,
    MockLibraryReportProcessor,
)


def fixture_logger_call_send(url, data):
    send_msg = "Sending HTTP Request to: {url}"
    if data:
        send_msg += "\n--Debug Input Start--\n{data}\n--Debug Input End--\n"
    return mock.call.debug(send_msg.format(url=url, data=data))


def fixture_logger_call_debug_data(url, data):
    send_msg = dedent(
        """\
        Communication debug info for calling: {url}
        --Debug Communication Info Start--
        {data}
        --Debug Communication Info End--
        """
    )
    return mock.call.debug(send_msg.format(url=url, data=data))


def fixture_logger_call_connected(url, response_code, response_data):
    result_msg = (
        "Finished calling: {url}\nResponse Code: {code}"
        + "\n--Debug Response Start--\n{response}\n--Debug Response End--\n"
    )
    return mock.call.debug(
        result_msg.format(url=url, code=response_code, response=response_data)
    )


def fixture_logger_call_not_connected(node, reason):
    msg = "Unable to connect to {node} ({reason})"
    return mock.call.debug(msg.format(node=node, reason=reason))


def fixture_logger_call_proxy_set():
    return mock.call.warning(
        "Proxy is set in environment variables, try disabling it"
    )


def fixture_logger_calls_on_success(
    url, response_code, response_data, debug_data
):
    return [
        fixture_logger_call_connected(url, response_code, response_data),
        fixture_logger_call_debug_data(url, debug_data),
    ]


def fixture_report_item_list_send(url, data):
    return [
        fixture.debug(
            report_codes.NODE_COMMUNICATION_STARTED, target=url, data=data
        )
    ]


def fixture_report_item_list_debug(url, data):
    return [
        fixture.debug(
            report_codes.NODE_COMMUNICATION_DEBUG_INFO, target=url, data=data
        )
    ]


def fixture_report_item_list_connected(url, response_code, response_data):
    return [
        fixture.debug(
            report_codes.NODE_COMMUNICATION_FINISHED,
            target=url,
            response_code=response_code,
            response_data=response_data,
        )
    ]


def fixture_report_item_list_not_connected(node, reason):
    return [
        fixture.debug(
            report_codes.NODE_COMMUNICATION_NOT_CONNECTED,
            node=node,
            reason=reason,
        )
    ]


def fixture_report_item_list_proxy_set(node, address):
    return [
        fixture.warn(
            report_codes.NODE_COMMUNICATION_PROXY_IS_SET,
            node=node,
            address=address,
        )
    ]


def fixture_report_item_list_on_success(
    url, response_code, response_data, debug_data
):
    return fixture_report_item_list_connected(
        url, response_code, response_data
    ) + fixture_report_item_list_debug(url, debug_data)


def fixture_request():
    return Request(RequestTarget("host"), RequestData("action"))


class CommunicatorLoggerTest(TestCase):
    def setUp(self):
        self.logger = mock.MagicMock(spec_set=logging.Logger)
        self.log_reporter = ReportProcessorToLog(self.logger)
        self.reporter = MockLibraryReportProcessor()
        self.com_logger = logger.CommunicatorLogger(
            [self.reporter, self.log_reporter]
        )

    def test_log_request_start(self):
        request = fixture_request()
        self.com_logger.log_request_start(request)
        self.reporter.assert_reports(
            fixture_report_item_list_send(request.url, request.data)
        )
        self.assertEqual(
            [fixture_logger_call_send(request.url, request.data)],
            self.logger.mock_calls,
        )

    def test_log_response_connected(self):
        expected_code = 200
        expected_data = "data"
        expected_debug_data = "* text\n>> data out\n"
        response = Response.connection_successful(
            MockCurlSimple(
                info={pycurl.RESPONSE_CODE: expected_code},
                output=expected_data.encode("utf-8"),
                debug_output=expected_debug_data.encode("utf-8"),
                request=fixture_request(),
            )
        )
        self.com_logger.log_response(response)
        self.reporter.assert_reports(
            fixture_report_item_list_on_success(
                response.request.url,
                expected_code,
                expected_data,
                expected_debug_data,
            )
        )
        logger_calls = fixture_logger_calls_on_success(
            response.request.url,
            expected_code,
            expected_data,
            expected_debug_data,
        )
        self.assertEqual(logger_calls, self.logger.mock_calls)

    @mock.patch("pcs.common.communication.logger.is_proxy_set")
    def test_log_response_not_connected(self, mock_proxy):
        mock_proxy.return_value = False
        expected_debug_data = "* text\n>> data out\n"
        error_msg = "error"
        response = Response.connection_failure(
            MockCurlSimple(
                debug_output=expected_debug_data.encode("utf-8"),
                request=fixture_request(),
            ),
            pycurl.E_HTTP_POST_ERROR,
            error_msg,
        )
        self.com_logger.log_response(response)
        self.reporter.assert_reports(
            fixture_report_item_list_not_connected(
                response.request.host_label, error_msg
            )
            + fixture_report_item_list_debug(
                response.request.url, expected_debug_data
            )
        )
        logger_calls = [
            fixture_logger_call_not_connected(
                response.request.host_label, error_msg
            ),
            fixture_logger_call_debug_data(
                response.request.url, expected_debug_data
            ),
        ]
        self.assertEqual(logger_calls, self.logger.mock_calls)

    @mock.patch("pcs.common.communication.logger.is_proxy_set")
    def test_log_response_not_connected_with_proxy(self, mock_proxy):
        mock_proxy.return_value = True
        expected_debug_data = "* text\n>> data out\n"
        error_msg = "error"
        response = Response.connection_failure(
            MockCurlSimple(
                debug_output=expected_debug_data.encode("utf-8"),
                request=fixture_request(),
            ),
            pycurl.E_HTTP_POST_ERROR,
            error_msg,
        )
        self.com_logger.log_response(response)
        self.reporter.assert_reports(
            fixture_report_item_list_not_connected(
                response.request.host_label, error_msg
            )
            + fixture_report_item_list_proxy_set(
                response.request.host_label, response.request.host_label
            )
            + fixture_report_item_list_debug(
                response.request.url, expected_debug_data
            )
        )
        logger_calls = [
            fixture_logger_call_not_connected(
                response.request.host_label, error_msg
            ),
            fixture_logger_call_proxy_set(),
            fixture_logger_call_debug_data(
                response.request.url, expected_debug_data
            ),
        ]
        self.assertEqual(logger_calls, self.logger.mock_calls)

    def test_log_retry(self):
        prev_addr = "addr"
        prev_port = 2225
        prev_host = Destination(prev_addr, prev_port)
        response = Response.connection_failure(
            MockCurlSimple(request=fixture_request()),
            pycurl.E_HTTP_POST_ERROR,
            "e",
        )
        self.com_logger.log_retry(response, prev_host)
        self.reporter.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_RETRYING,
                    node=response.request.host_label,
                    failed_address=prev_addr,
                    failed_port=str(prev_port),
                    next_address=response.request.dest.addr,
                    next_port=str(settings.pcsd_default_port),
                    request=response.request.url,
                )
            ]
        )
        logger_call = mock.call.warning(
            (
                "Unable to connect to '{label}' via address '{old_addr}' and "
                "port '{old_port}'. Retrying request '{req}' via address "
                "'{new_addr}' and port '{new_port}'"
            ).format(
                label=response.request.host_label,
                old_addr=prev_addr,
                old_port=prev_port,
                new_addr=response.request.dest.addr,
                new_port=settings.pcsd_default_port,
                req=response.request.url,
            )
        )
        self.assertEqual([logger_call], self.logger.mock_calls)

    def test_log_no_more_addresses(self):
        response = Response.connection_failure(
            MockCurlSimple(request=fixture_request()),
            pycurl.E_HTTP_POST_ERROR,
            "e",
        )
        self.com_logger.log_no_more_addresses(response)
        self.reporter.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_NO_MORE_ADDRESSES,
                    node=response.request.host_label,
                    request=response.request.url,
                )
            ]
        )
        logger_call = mock.call.warning(
            "Unable to connect to '{label}' via any of its addresses".format(
                label=response.request.host_label
            )
        )
        self.assertEqual([logger_call], self.logger.mock_calls)



from typing import (
    Optional,
    TypedDict,
)

from tornado.web import RequestHandler

from pcs.daemon.app.auth import UnixSocketAuthProvider
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from .session import (
    Session,
    Storage,
)

PCSD_SESSION = "pcsd.sid"


class SessionAuthProvider(UnixSocketAuthProvider):
    class CookieOptions(TypedDict):
        secure: bool
        httponly: bool
        samesite: str

    __cookie_options: CookieOptions = {
        "secure": True,
        "httponly": True,
        # rhbz#2097393
        # Prevent a cookie to be sent on cross-site requests, allow it to be
        # sent when navigating to pcs web UI from an external site.
        "samesite": "Lax",
    }

    def __init__(
        self,
        handler: RequestHandler,
        auth_provider: AuthProvider,
        session_storage: Storage,
    ) -> None:
        super().__init__(handler, auth_provider)
        self.__session_storage = session_storage
        self.__session: Optional[Session] = None

    def init_session(self) -> None:
        self.__session = self._get_session()

    async def auth_by_sid(self) -> AuthUser:
        if self.__session:
            return await self.login_user(self.__session.username)
        return await self.auth_by_socket_user()

    def _get_session(self) -> Optional[Session]:
        if self.__sid_from_client is None:
            return None
        return self.__session_storage.get(self.__sid_from_client)

    def session_logout(self) -> None:
        if self.__session is not None:
            self.__session_storage.destroy(self.__session.sid)
        self.__session = None
        self._sid_to_cookies()

    def _sid_to_cookies(self) -> None:
        """
        Write the session id into a response cookie.
        """
        if self.__session:
            self._handler.set_cookie(
                PCSD_SESSION, self.__session.sid, **self.__cookie_options
            )
        else:
            self._handler.clear_cookie(PCSD_SESSION)

    def put_request_cookies_sid_to_response_cookies_sid(self) -> None:
        """
        If sid came in the request cookies put it into response cookies. But do
        not start new one.
        """
        # TODO this method should exist temporarily (for sinatra compatibility)
        if self.__sid_from_client is not None:
            self._handler.set_cookie(
                PCSD_SESSION, self.__sid_from_client, **self.__cookie_options
            )

    def is_sid_in_request_cookies(self) -> bool:
        return self.__sid_from_client is not None

    @property
    def __sid_from_client(self) -> Optional[str]:
        return self._handler.get_cookie(PCSD_SESSION, default=None)

    def update_session(self, auth_user: Optional[AuthUser]) -> None:
        if auth_user:
            if (
                not self.__session
                or self.__session.username != auth_user.username
            ):
                self.__session = self.__session_storage.login(
                    auth_user.username
                )
        else:
            self.__session = None
        self._sid_to_cookies()



import difflib
import doctest
import re

from lxml.doctestcompare import LXMLOutputChecker
from lxml.etree import LXML_VERSION

from pcs.common import reports
from pcs.lib.errors import LibraryError

from pcs_test.tools.fixture import ReportItemFixture


def prepare_diff(first, second):
    """
    Return a string containing a diff of first and second
    """
    return "".join(
        difflib.Differ().compare(first.splitlines(1), second.splitlines(1))
    )


def ac(a, b):
    # pylint: disable=invalid-name
    """
    Compare the actual output 'a' and an expected output 'b', print diff b a
    """
    if a != b:
        raise AssertionError(
            "strings not equal:\n{0}".format(prepare_diff(b, a))
        )


def start_tag_error_text():
    """lxml 3.7+ gives a longer 'start tag expected' error message,
    handle it here so multiple tests can just get the appropriate
    string from this function.
    """
    msg = "Start tag expected, '<' not found, line 1, column 1"
    if LXML_VERSION >= (3, 7, 0, 0):
        msg += " (<string>, line 1)"
    return msg


class AssertPcsMixin:
    """Run pcs command and assert its result"""

    def assert_pcs_success_all(self, command_list):
        for command in command_list:
            stdout, stderr, retval = self.pcs_runner.run(command)
            if retval != 0:
                raise AssertionError(
                    f"Command '{command}' did not succeed\n"
                    f"** return code: {retval}\n"
                    f"** stdout:\n{stdout}\n"
                    f"** stderr:\n{stderr}\n"
                )

    def assert_pcs_success(
        self,
        command,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
        stderr_full=None,
        stderr_start=None,
        stderr_regexp=None,
        despace=False,
    ):
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments
        # It is common that successful commands don't print anything, so we
        # default stdout and stderr to an empty string if not specified
        # otherwise.
        self.assert_pcs_result(
            command,
            stdout_full=self.__default_output_to_empty_str(
                stdout_full, stdout_start, stdout_regexp
            ),
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            stderr_full=self.__default_output_to_empty_str(
                stderr_full, stderr_start, stderr_regexp
            ),
            stderr_start=stderr_start,
            stderr_regexp=stderr_regexp,
            returncode=0,
            despace=despace,
        )

    def assert_pcs_fail(
        self,
        command,
        stderr_full=None,
        stderr_start=None,
        stderr_regexp=None,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
    ):
        # It is common that failed commands don't print anything to stdout, so
        # we default stdout to an empty string if not specified otherwise.
        self.assert_pcs_result(
            command,
            stdout_full=self.__default_output_to_empty_str(
                stdout_full, stdout_start, stdout_regexp
            ),
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            stderr_full=stderr_full,
            stderr_start=stderr_start,
            stderr_regexp=stderr_regexp,
            returncode=1,
        )

    def assert_pcs_fail_regardless_of_force(
        self,
        command,
        stderr_full=None,
        stderr_start=None,
        stderr_regexp=None,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
    ):
        self.assert_pcs_fail(
            command,
            stdout_full=stdout_full,
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            stderr_full=stderr_full,
            stderr_start=stderr_start,
            stderr_regexp=stderr_regexp,
        )
        self.assert_pcs_fail(
            command + ["--force"],
            stdout_full=stdout_full,
            stdout_start=stdout_start,
            stdout_regexp=stdout_regexp,
            stderr_full=stderr_full,
            stderr_start=stderr_start,
            stderr_regexp=stderr_regexp,
        )

    def assert_pcs_result(  # noqa: PLR0913
        self,
        command,
        *,
        stdout_full=None,
        stdout_start=None,
        stdout_regexp=None,
        stderr_full=None,
        stderr_start=None,
        stderr_regexp=None,
        returncode=0,
        despace=False,
    ):
        # pylint: disable=too-many-arguments
        self.__check_output_specified(
            stdout_full, stdout_start, stdout_regexp, "stdout"
        )
        self.__check_output_specified(
            stderr_full, stderr_start, stderr_regexp, "stderr"
        )

        stdout_actual, stderr_actual = self.assert_pcs_success_ignore_output(
            command, returncode
        )

        message_template = (
            f"{{reason}}\n** Command: {command}\n** {{detail}}\n"
            f"** Retval: {returncode}\n"
            f"** Full stdout:\n{stdout_actual}\n"
            f"** Full stderr:\n{stderr_actual}\n"
        )
        self.__fail_on_unexpected_output(
            stdout_actual,
            stdout_full,
            stdout_start,
            stdout_regexp,
            despace,
            "stdout",
            message_template,
        )

        self.__fail_on_unexpected_output(
            stderr_actual,
            stderr_full,
            stderr_start,
            stderr_regexp,
            despace,
            "stderr",
            message_template,
        )

    def assert_pcs_success_ignore_output(self, command, returncode=0):
        stdout, stderr, retval_actual = self.pcs_runner.run(command)

        self.assertEqual(
            returncode,
            retval_actual,
            (
                f"Expected return code '{returncode}' but was '{retval_actual}'\n"
                f"** command: {command}\n"
                f"** stdout:\n{stdout}\n"
                f"** stderr:\n{stderr}\n"
            ),
        )
        return stdout, stderr

    @staticmethod
    def __default_output_to_empty_str(output_full, output_start, output_regexp):
        if (
            output_start is None
            and output_full is None
            and output_regexp is None
        ):
            return ""
        return output_full

    @staticmethod
    def __check_output_specified(
        output_full, output_start, output_regexp, label
    ):
        msg = (
            f"Please specify exactly one: {label}_full or {label}_start or "
            f"{label}_regexp"
        )
        specified_output = [
            output
            for output in (output_full, output_start, output_regexp)
            if output is not None
        ]
        if not specified_output:
            raise TypeError(msg + ", none specified")
        if len(specified_output) > 1:
            raise TypeError(msg + ", more than one specified")

    def __fail_on_unexpected_output(
        self,
        output_actual,
        output_full,
        output_start,
        output_regexp,
        despace,
        label,
        message_template,
    ):
        if output_start:
            expected_start = self.__prepare_output(output_start)
            if not output_actual.startswith(expected_start):
                diff = prepare_diff(
                    output_actual[: len(expected_start)], expected_start
                )
                raise AssertionError(
                    message_template.format(
                        reason=f"{label} does not start as expected",
                        detail=f"diff is (expected is 2nd):\n{diff}\n",
                    )
                )
            return

        if output_regexp:
            if not isinstance(output_regexp, re.Pattern):
                output_regexp = re.compile(output_regexp)
            if not output_regexp.search(output_actual):
                flags = ", ".join(
                    self.__prepare_regexp_flags(output_regexp.flags)
                )
                raise AssertionError(
                    message_template.format(
                        reason=f"{label} does not match the expected regexp",
                        detail=(
                            f"regexp:\n{output_regexp.pattern}\n"
                            f"regexp flags: {flags}\n"
                        ),
                    )
                )
            return

        expected_full = self.__prepare_output(output_full)
        if (despace and _despace(output_actual) != _despace(expected_full)) or (
            not despace and output_actual != expected_full
        ):
            diff = prepare_diff(output_actual, expected_full)
            raise AssertionError(
                message_template.format(
                    reason=f"{label} is not as expected",
                    detail=f"diff is (expected is 2nd):\n{diff}\n",
                ),
            )

    @staticmethod
    def __prepare_output(output):
        return "\n".join(output + [""]) if isinstance(output, list) else output

    @staticmethod
    def __prepare_regexp_flags(used_flags):
        return sorted([flag.name for flag in re.RegexFlag if used_flags & flag])


class ExtendedAssertionsMixin:
    def assert_raises(
        self, expected_exception, callable_obj, property_dict=None
    ):
        if property_dict is None:
            property_dict = {}
        try:
            callable_obj()
            raise AssertionError(
                "No exception raised. Expected exception: {exception}".format(
                    exception=expected_exception.__class__.__name__
                )
            )
        except expected_exception as e:
            for prop, value in property_dict.items():
                try:
                    self.assertEqual(value, getattr(e, prop))
                except AttributeError as exc:
                    raise AssertionError(
                        "Property {property} doesn't exist in exception"
                        " {exception}".format(
                            property=prop, exception=exc.__class__.__name__
                        )
                    ) from exc


def assert_xml_equal(expected_xml, got_xml, context_explanation=""):
    checker = LXMLOutputChecker()
    if not checker.check_output(expected_xml, got_xml, 0):
        raise AssertionError(
            "{context_explanation}{xml_diff}".format(
                context_explanation=(
                    ""
                    if not context_explanation
                    else "\n{0}\n".format(context_explanation)
                ),
                xml_diff=checker.output_difference(
                    doctest.Example("", expected_xml), got_xml, 0
                ),
            )
        )


SEVERITY_SHORTCUTS = {
    reports.ReportItemSeverity.INFO: "I",
    reports.ReportItemSeverity.WARNING: "W",
    reports.ReportItemSeverity.DEPRECATION: "DW",
    reports.ReportItemSeverity.ERROR: "E",
    reports.ReportItemSeverity.DEBUG: "D",
}


def _format_report_item_info(info):
    return ", ".join(
        ["{0}:{1}".format(key, repr(value)) for key, value in info.items()]
    )


def _expected_report_item_format(report_item_expectation):
    return "{0} {1} {{{2}}} ! {3} {4}".format(
        SEVERITY_SHORTCUTS.get(
            report_item_expectation[0], report_item_expectation[0]
        ),
        report_item_expectation[1],
        _format_report_item_info(report_item_expectation[2]),
        (
            report_item_expectation[3]
            if len(report_item_expectation) > 3
            else None
        ),
        (
            report_item_expectation[4]
            if len(report_item_expectation) > 4
            else None
        ),
    )


def _format_report_item(report_item):
    return _expected_report_item_format(
        (
            report_item.severity.level,
            report_item.message.code,
            report_item.message.to_dto().payload,
            report_item.severity.force_code,
            report_item.context,
        )
    )


def assert_report_item_equal(real_report_item, report_item_info):
    if not __report_item_equal(real_report_item, report_item_info):
        raise AssertionError(
            "ReportItem not equal\nexpected: {0}\nactual:   {1}".format(
                repr(
                    (
                        report_item_info[0],
                        report_item_info[1],
                        report_item_info[2],
                        (
                            None
                            if len(report_item_info) < 4
                            else report_item_info[3]
                        ),
                        (
                            None
                            if len(report_item_info) < 5
                            else report_item_info[4]
                        ),
                    )
                ),
                _format_report_item(real_report_item),
            )
        )


def _unexpected_report_given(
    remaining_expected_report_info_list,
    expected_report_info_list,
    real_report_item,
    real_report_item_list,
):
    return AssertionError(
        (
            "\n  Unexpected real report given:"
            "\n  =============================\n    {0}\n"
            "\n  remaining expected reports ({1}) are:"
            "\n  ------------------------------------\n    {2}\n"
            "\n  all expected reports ({3}) are:"
            "\n  ------------------------------\n    {4}\n"
            "\n  all real reports ({5}):"
            "\n  ---------------------\n    {6}"
        ).format(
            _format_report_item(real_report_item),
            len(remaining_expected_report_info_list),
            (
                "\n    ".join(
                    map(
                        _expected_report_item_format,
                        remaining_expected_report_info_list,
                    )
                )
                if remaining_expected_report_info_list
                else "No other report is expected!"
            ),
            len(expected_report_info_list),
            (
                "\n    ".join(
                    map(
                        _expected_report_item_format,
                        expected_report_info_list,
                    )
                )
                if expected_report_info_list
                else "No report is expected!"
            ),
            len(real_report_item_list),
            "\n    ".join(map(_format_report_item, real_report_item_list)),
        )
    )


def assert_report_item_list_equal(
    real_report_item_list, expected_report_info_list, hint=""
):
    remaining_expected_report_info_list = expected_report_info_list[:]
    duplicate_report_item_is_missing = False
    for real_report_item in real_report_item_list:
        found_report_info = __find_report_info(
            expected_report_info_list, real_report_item
        )
        if found_report_info is None:
            if (
                real_report_item.severity.level
                == reports.ReportItemSeverity.DEBUG
            ):
                # ignore debug report items not specified as expected
                continue
            raise _unexpected_report_given(
                remaining_expected_report_info_list,
                expected_report_info_list,
                real_report_item,
                real_report_item_list,
            )
        if found_report_info in remaining_expected_report_info_list:
            remaining_expected_report_info_list.remove(found_report_info)
        else:
            duplicate_report_item_is_missing = True
    if remaining_expected_report_info_list or duplicate_report_item_is_missing:

        def format_items(item_type, item_list):
            caption = "{0} ReportItems({1})".format(item_type, len(item_list))
            return "{0}\n{1}\n{2}".format(
                caption, "-" * len(caption), "\n".join(map(repr, item_list))
            )

        raise AssertionError(
            "\nReport lists doesn't match{0}\n\n{1}\n\n{2}".format(
                "\n{0}".format(hint) if hint else "",
                format_items("expected", expected_report_info_list),
                format_items("real", real_report_item_list),
            )
        )


def assert_raise_library_error(callable_obj, *report_info_list):
    try:
        callable_obj()
        raise AssertionError("LibraryError not raised")
    except LibraryError as e:
        assert_report_item_list_equal(e.args, list(report_info_list))


def __find_report_info(expected_report_info_list, real_report_item):
    for report_info in expected_report_info_list:
        if __report_item_equal(real_report_item, report_info):
            return report_info
    return None


def __report_item_equal(
    real_report_item: reports.ReportItem, report_item_info: ReportItemFixture
) -> bool:
    report_dto: reports.ReportItemDto = real_report_item.to_dto()
    return (
        report_dto.severity.level == report_item_info[0]
        and report_dto.message.code == report_item_info[1]
        and report_dto.message.payload == report_item_info[2]
        and (
            report_dto.severity.force_code
            == (None if len(report_item_info) < 4 else report_item_info[3])
        )
        and report_dto.context
        == (report_item_info[4] if len(report_item_info) >= 5 else None)
    )


def assert_pcs_status(status1, status2):
    if _despace(status1) != _despace(status2):
        raise AssertionError(
            "strings not equal:\n{0}".format(prepare_diff(status1, status2))
        )


def _despace(string):
    # ignore whitespace changes between various pacemaker versions
    return re.sub(r"[ \t]+", " ", string)



from dataclasses import replace
from typing import Sequence

from pcs.common.host import PcsKnownHost
from pcs.common.types import StringSequence
from pcs.lib.interface.config import SyncVersionFacadeInterface

from .types import KnownHosts


class Facade(SyncVersionFacadeInterface):
    def __init__(self, parsed_config: KnownHosts):
        super().__init__(parsed_config)

    @classmethod
    def create(cls, data_version: int = 1) -> "Facade":
        return cls(
            KnownHosts(
                format_version=1, data_version=data_version, known_hosts={}
            )
        )

    @property
    def config(self) -> KnownHosts:
        return self._config

    @property
    def data_version(self) -> int:
        return self.config.data_version

    def set_data_version(self, new_version: int) -> None:
        self._set_config(replace(self.config, data_version=new_version))

    @property
    def known_hosts(self) -> dict[str, PcsKnownHost]:
        return dict(self.config.known_hosts)

    def update_known_hosts(self, hosts: Sequence[PcsKnownHost]) -> None:
        updated_hosts = self.known_hosts
        for new_host in hosts:
            updated_hosts[new_host.name] = new_host

        self._set_config(replace(self.config, known_hosts=updated_hosts))

    def remove_known_hosts(self, hosts: StringSequence) -> None:
        updated_hosts = self.known_hosts
        for host_name in hosts:
            if host_name in updated_hosts:
                del updated_hosts[host_name]

        self._set_config(replace(self.config, known_hosts=updated_hosts))



from typing import cast


class CibResourceDiscovery(str):
    __slots__ = ()
    ALWAYS = cast("CibResourceDiscovery", "always")
    NEVER = cast("CibResourceDiscovery", "never")
    EXCLUSIVE = cast("CibResourceDiscovery", "exclusive")


class CibResourceSetOrdering(str):
    __slots__ = ()
    GROUP = cast("CibResourceSetOrdering", "group")
    LISTED = cast("CibResourceSetOrdering", "listed")


class CibResourceSetOrderType(str):
    __slots__ = ()
    OPTIONAL = cast("CibResourceSetOrderType", "Optional")
    MANDATORY = cast("CibResourceSetOrderType", "Mandatory")
    SERIALIZE = cast("CibResourceSetOrderType", "Serialize")


class CibTicketLossPolicy(str):
    __slots__ = ()
    STOP = cast("CibTicketLossPolicy", "stop")
    DEMOTE = cast("CibTicketLossPolicy", "demote")
    FENCE = cast("CibTicketLossPolicy", "fence")
    FREEZE = cast("CibTicketLossPolicy", "freeze")



from typing import (
    Optional,
    cast,
)

from lxml.etree import _Element

from pcs.lib.booth.constants import DEFAULT_INSTANCE_NAME

_BOOTH_ATTRIBUTE = "booth-cfg-name"


def get_booth_ticket_names(
    cib: _Element, booth_instance: Optional[str] = None
) -> list[str]:
    """
    Return names of booth tickets present in CIB for the specified booth
    instance

    cib -- element representing the CIB
    booth_instance -- name of the booth instance, default instance name is used
        if name is not provided
    """
    return cast(
        list[str],
        cib.xpath(
            f"status/tickets/ticket_state[@{_BOOTH_ATTRIBUTE}=$instance]/@id",
            instance=booth_instance or DEFAULT_INSTANCE_NAME,
        ),
    )


def get_ticket_names(cib: _Element) -> list[str]:
    """
    Return names of all tickets present in CIB

    cib -- element representing the CIB
    """
    return cast(list[str], cib.xpath("status/tickets/ticket_state/@id"))



from functools import partial

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.pacemaker.constraint import (
    CibConstraintColocationAttributesDto,
    CibConstraintColocationDto,
    CibConstraintColocationSetDto,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.const import TAG_CONSTRAINT_COLOCATION as TAG
from pcs.lib.cib.constraint import constraint
from pcs.lib.cib.constraint.common import is_set_constraint
from pcs.lib.cib.constraint.resource_set import (
    constraint_element_to_resource_set_dto_list,
)
from pcs.lib.cib.rule import RuleInEffectEval
from pcs.lib.cib.rule.cib_to_dto import rule_element_to_dto
from pcs.lib.cib.tools import (
    check_new_id_applicable,
    role_constructor,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import (
    SCORE_INFINITY,
    is_score,
)
from pcs.lib.tools import get_optional_value

_DESCRIPTION = "constraint id"


def is_colocation_constraint(element: _Element) -> bool:
    return element.tag == TAG


def prepare_options_with_set(cib, options, resource_set_list):
    options = constraint.prepare_options(
        ("score",),
        options,
        partial(constraint.create_id, cib, "colocation", resource_set_list),
        partial(check_new_id_applicable, cib, _DESCRIPTION),
    )

    if "score" in options:
        if not is_score(options["score"]):
            raise LibraryError(
                ReportItem.error(
                    reports.messages.InvalidScore(options["score"])
                )
            )
    else:
        options["score"] = SCORE_INFINITY

    return options


def _element_to_attributes_dto(
    element: _Element, rule_in_effect_eval: RuleInEffectEval
) -> CibConstraintColocationAttributesDto:
    return CibConstraintColocationAttributesDto(
        constraint_id=str(element.attrib["id"]),
        score=element.get("score"),
        influence=element.get("influence"),
        lifetime=[
            rule_element_to_dto(rule_in_effect_eval, rule_el)
            for rule_el in element.findall("./lifetime/rule")
        ],
    )


def _constraint_el_to_dto(
    element: _Element, rule_in_effect_eval: RuleInEffectEval
) -> CibConstraintColocationDto:
    return CibConstraintColocationDto(
        resource_id=str(element.attrib["rsc"]),
        with_resource_id=str(element.attrib["with-rsc"]),
        node_attribute=element.get("node-attribute"),
        resource_role=get_optional_value(
            role_constructor, element.get("rsc-role")
        ),
        with_resource_role=get_optional_value(
            role_constructor, element.get("with-rsc-role")
        ),
        resource_instance=get_optional_value(int, element.get("rsc-instance")),
        with_resource_instance=get_optional_value(
            int, element.get("with-rsc-instance")
        ),
        attributes=_element_to_attributes_dto(element, rule_in_effect_eval),
    )


def _set_constraint_el_to_dto(
    element: _Element, rule_in_effect_eval: RuleInEffectEval
) -> CibConstraintColocationSetDto:
    return CibConstraintColocationSetDto(
        resource_sets=constraint_element_to_resource_set_dto_list(element),
        attributes=_element_to_attributes_dto(element, rule_in_effect_eval),
    )


def get_all_as_dtos(
    constraints_el: _Element, rule_in_effect_eval: RuleInEffectEval
) -> tuple[
    list[CibConstraintColocationDto], list[CibConstraintColocationSetDto]
]:
    plain_list: list[CibConstraintColocationDto] = []
    set_list: list[CibConstraintColocationSetDto] = []
    for constraint_el in constraints_el.findall(f"./{TAG}"):
        if is_set_constraint(constraint_el):
            set_list.append(
                _set_constraint_el_to_dto(constraint_el, rule_in_effect_eval)
            )
        else:
            plain_list.append(
                _constraint_el_to_dto(constraint_el, rule_in_effect_eval)
            )
    return plain_list, set_list



import json
import shlex
from unittest import TestCase

from lxml import etree

from pcs.common.interface import dto
from pcs.common.pacemaker.node import CibNodeDto, CibNodeListDto
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto

from pcs_test.tier1.legacy.common import (
    FIXTURE_UTILIZATION_WARNING,
)
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

FIXTURE_EMPTY_NODES_DTO = CibNodeListDto(
    nodes=[
        CibNodeDto(
            id="1",
            uname="rh7-1",
            description=None,
            score=None,
            type=None,
            instance_attributes=[],
            utilization=[],
        ),
        CibNodeDto(
            id="2",
            uname="rh7-2",
            description=None,
            score=None,
            type=None,
            instance_attributes=[],
            utilization=[],
        ),
    ]
)

FIXTURE_NODES_DTO = CibNodeListDto(
    nodes=[
        CibNodeDto(
            id="1",
            uname="rh7-1",
            description=None,
            score=None,
            type=None,
            instance_attributes=[
                CibNvsetDto(
                    id="nodes-1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="nodes-1-a", name="a", value="1"),
                        CibNvpairDto(id="nodes-1-b", name="b", value="2"),
                    ],
                )
            ],
            utilization=[
                CibNvsetDto(
                    id="nodes-1-utilization",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(
                            id="nodes-1-utilization-cpu", name="cpu", value="4"
                        ),
                        CibNvpairDto(
                            id="nodes-1-utilization-ram", name="ram", value="32"
                        ),
                    ],
                )
            ],
        ),
        CibNodeDto(
            id="2",
            uname="rh7-2",
            description=None,
            score=None,
            type=None,
            instance_attributes=[
                CibNvsetDto(
                    id="nodes-2",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="nodes-2-a", name="a", value="1"),
                        CibNvpairDto(id="nodes-2-b", name="b", value="2"),
                    ],
                )
            ],
            utilization=[
                CibNvsetDto(
                    id="nodes-2-utilization",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(
                            id="nodes-2-utilization-cpu", name="cpu", value="8"
                        ),
                        CibNvpairDto(
                            id="nodes-2-utilization-ram", name="ram", value="64"
                        ),
                    ],
                )
            ],
        ),
    ]
)

FIXTURE_ATTRIBUTES_TEXT = """\
Node: rh7-1
  Attributes: nodes-1
    a=1
    b=2
Node: rh7-2
  Attributes: nodes-2
    a=1
    b=2
"""

FIXTURE_UTILIZATION_TEXT = """\
Node: rh7-1
  Utilization: nodes-1-utilization
    cpu=4
    ram=32
Node: rh7-2
  Utilization: nodes-2-utilization
    cpu=8
    ram=64
"""


class NodeOutputFormatJsonMixin:
    def setUp(self):
        self.maxDiff = None

    def test_empty_nodes(self):
        pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-empty-withnodes.xml")
        )
        stdout, stderr, retval = pcs_runner.run(
            ["node", self.command, "--output-format=json"]
        )
        self.assertEqual(
            json.loads(stdout), dto.to_dict(FIXTURE_EMPTY_NODES_DTO)
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        temp_cib = get_tmp_file("tier1_nodes_json")
        write_file_to_tmpfile(get_test_resource("cib-all.xml"), temp_cib)
        pcs_runner = PcsRunner(cib_file=temp_cib.name)
        stdout, stderr, retval = pcs_runner.run(
            ["node", self.command, "--output-format=json"]
        )
        self.assertEqual(json.loads(stdout), dto.to_dict(FIXTURE_NODES_DTO))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        temp_cib.close()


class NodeAttributeOutputFormatJson(NodeOutputFormatJsonMixin, TestCase):
    command = "attribute"


class NodeUtilizationOutputFormatJson(NodeOutputFormatJsonMixin, TestCase):
    command = "utilization"


class NodeAttributeUtilizationOutputFormatCmd(TestCase):
    def setUp(self):
        self.old_cib = get_tmp_file("tier1_node_cmd_old")
        write_file_to_tmpfile(get_test_resource("cib-all.xml"), self.old_cib)
        self.new_cib = get_tmp_file("tier1_node_cmd_new")
        write_file_to_tmpfile(
            get_test_resource("cib-empty-withnodes.xml"), self.new_cib
        )
        self.old_pcs_runner = PcsRunner(self.old_cib.name)
        self.new_pcs_runner = PcsRunner(self.new_cib.name)
        self.maxDiff = None

    def tearDown(self):
        self.old_cib.close()
        self.new_cib.close()

    def _get_as_json(self, runner, command):
        stdout, stderr, retval = runner.run(
            ["node", command, "--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_success(self):
        stdout_attribute, stderr, retval = self.old_pcs_runner.run(
            ["node", "attribute", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        stdout_utilization, stderr, retval = self.old_pcs_runner.run(
            ["node", "utilization", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        stdout = ";\n".join([stdout_attribute, stdout_utilization])
        commands = [
            shlex.split(command)[1:]
            for command in stdout.replace("\\n", "").strip().split(";\n")
        ]
        for cmd in commands:
            stdout, stderr, retval = self.new_pcs_runner.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(self.new_pcs_runner, "attribute"),
            self._get_as_json(self.old_pcs_runner, "attribute"),
        )
        self.assertEqual(
            self._get_as_json(self.new_pcs_runner, "utilization"),
            self._get_as_json(self.old_pcs_runner, "utilization"),
        )


class NodeOutputFormatTextMixin:
    def setUp(self):
        self.maxDiff = None

    def test_empty(self):
        pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-empty-withnodes.xml")
        )
        stdout, stderr, retval = pcs_runner.run(
            ["node", self.command, "--output-format=text"]
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, self.expected_stderr)
        self.assertEqual(retval, 0)

    def test_success(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-all.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["node", self.command, "--output-format=text"]
        )
        self.assertEqual(stdout, self.expected_stdout)
        self.assertEqual(stderr, self.expected_stderr)
        self.assertEqual(retval, 0)


class NodeAttributeOutputFormatText(NodeOutputFormatTextMixin, TestCase):
    command = "attribute"
    expected_stdout = FIXTURE_ATTRIBUTES_TEXT
    expected_stderr = ""


class NodeUtilizationOutputFormatText(NodeOutputFormatTextMixin, TestCase):
    command = "utilization"
    expected_stdout = FIXTURE_UTILIZATION_TEXT
    expected_stderr = FIXTURE_UTILIZATION_WARNING


class NodeOutputFormatErrorsMixin(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).findall(".//nodes")[0])
    )
):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_tag")
        write_file_to_tmpfile(
            get_test_resource("cib-empty-withnodes.xml"), self.temp_cib
        )
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_name_with_output_format_no_text(self):
        self.assert_pcs_fail(
            f"node {self.command} --output-format=json --name cpu".split(),
            "Error: filtering is not supported with --output-format=cmd|json\n",
        )

    def test_output_format_with_args(self):
        self.assert_pcs_fail(
            f"node {self.command} --output-format=cmd rh7-1 cpu=".split(),
            stderr_start="\nUsage: pcs node <command>",
        )

    def test_multiple_args(self):
        self.assert_pcs_fail(
            f"node {self.command} rh7-1 cpu".split(),
            "{warn}Error: missing value of 'cpu' option\n".format(
                warn=(
                    ""
                    if self.command != "utilization"
                    else FIXTURE_UTILIZATION_WARNING
                )
            ),
        )


class NodeAttributeOutputFormatErrors(NodeOutputFormatErrorsMixin, TestCase):
    command = "attribute"


class NodeUtilizationOutputFormatErrors(NodeOutputFormatErrorsMixin, TestCase):
    command = "utilization"



from unittest import TestCase, mock

from pcs import settings
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.resource_agent.dto import ResourceAgentParameterDto
from pcs.lib.commands import cluster_property

from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from .crm_attribute_mixins import CrmAttributeMetadataErrorMixin

READONLY_PROPERTIES = [
    "cluster-infrastructure",
    "cluster-name",
    "dc-version",
    "have-watchdog",
    "last-lrm-refresh",
]


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_rng/api/api-result.rng"),
)
class TestGetPropertiesMetadataCrmAttribute(
    CrmAttributeMetadataErrorMixin, TestCase
):
    _load_cib_when_metadata_error = False

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def metadata_error_command(self):
        return self.command()

    def command(self):
        return cluster_property.get_properties_metadata(
            self.env_assist.get_env()
        )

    def _load_fake_agent_test_metadata(self):
        self.config.runner.pcmk.is_crm_attribute_list_options_supported(
            is_supported=True
        )
        self.config.runner.pcmk.load_crm_attribute_metadata(
            agent_name="cluster-options",
            stdout="""
                <pacemaker-result api-version="2.38" request="crm_attribute --list-options=cluster --output-as xml">
                  <resource-agent name="cluster-options" version="2.1.5-7.el9">
                    <version>1.1</version>
                    <longdesc lang="en">agent longdesc</longdesc>
                    <shortdesc lang="en">agent shortdesc</shortdesc>
                    <parameters>
                      <parameter name="property-name" advanced="0" generated="0">
                        <longdesc lang="en">longdesc</longdesc>
                        <shortdesc lang="en">shortdesc</shortdesc>
                        <content type="boolean" default="false"/>
                      </parameter>
                      <parameter name="enum-property" advanced="0" generated="0">
                        <longdesc lang="en">same desc</longdesc>
                        <shortdesc lang="en">same desc</shortdesc>
                        <content type="select" default="stop">
                          <option value="stop" />
                          <option value="freeze" />
                          <option value="ignore" />
                          <option value="demote" />
                          <option value="suicide" />
                        </content>
                      </parameter>
                      <parameter name="advanced-property" advanced="0" generated="0">
                        <longdesc lang="en">longdesc</longdesc>
                        <shortdesc lang="en">
                          *** Advanced Use Only *** advanced shortdesc
                        </shortdesc>
                        <content type="boolean" default="false"/>
                      </parameter>
                    </parameters>
                  </resource-agent>
                  <status code="0" message="OK"/>
                </pacemaker-result>
            """,
        )

    def test_get_properties_metadata(self):
        self._load_fake_agent_test_metadata()
        self.assertEqual(
            self.command(),
            ClusterPropertyMetadataDto(
                properties_metadata=[
                    ResourceAgentParameterDto(
                        name="property-name",
                        shortdesc="shortdesc",
                        longdesc="shortdesc.\nlongdesc",
                        type="boolean",
                        default="false",
                        enum_values=None,
                        required=False,
                        advanced=False,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=False,
                    ),
                    ResourceAgentParameterDto(
                        name="enum-property",
                        shortdesc="same desc",
                        longdesc=None,
                        type="select",
                        default="stop",
                        enum_values=[
                            "stop",
                            "freeze",
                            "ignore",
                            "demote",
                            "suicide",
                        ],
                        required=False,
                        advanced=False,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=False,
                    ),
                    ResourceAgentParameterDto(
                        name="advanced-property",
                        shortdesc="advanced shortdesc",
                        longdesc="advanced shortdesc.\nlongdesc",
                        type="boolean",
                        default="false",
                        enum_values=None,
                        required=False,
                        advanced=True,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=False,
                    ),
                ],
                readonly_properties=READONLY_PROPERTIES,
            ),
        )
        self.env_assist.assert_reports([])
